# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- simpler transformations
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>,
#            Lele Gaifax <lele@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
from functools import reduce

from . import _normalize_name, _normalize_dict_keys

from ..js_ast import (
    JSAssignmentExpression,
    JSArrowFunction,
    JSAttribute,
    JSAugAssignStatement,
    JSAwait,
    JSBinOp,
    JSBreakStatement,
    JSCall,
    JSContinueStatement,
    JSDeleteStatement,
    JSDict,
    JSExport,
    JSExpressionStatement,
    JSFalse,
    JSIfExp,
    JSIfStatement,
    JSList,
    JSName,
    JSNull,
    JSNum,
    JSOpAdd,
    JSOpAnd,
    JSOpBitAnd,
    JSOpBitOr,
    JSOpBitXor,
    JSOpDiv,
    JSOpGt,
    JSOpGtE,
    JSOpIn,
    JSOpInvert,
    JSOpLShift,
    JSOpLt,
    JSOpLtE,
    JSOpMod,
    JSOpMult,
    JSOpNot,
    JSOpOr,
    JSOpRShift,
    JSOpSub,
    JSOpUSub,
    JSPass,
    JSRest,
    JSReturnStatement,
    JSStatements,
    JSStr,
    JSSubscript,
    JSTemplateLiteral,
    JSTrue,
    JSUnaryOp,
    JSWhileStatement,
    JSYield,
    JSYieldStar,
)


#### Statements


def Assign_default(t, x):
    y = JSAssignmentExpression(x.targets[-1], x.value)
    for i in range(len(x.targets) - 1):
        y = JSAssignmentExpression(x.targets[-(2 + i)], y)
    return JSExpressionStatement(y)


def Assign_all(t, x):
    if len(x.targets) == 1 and isinstance(x.targets[0], ast.Name) and \
       x.targets[0].id == '__all__':
        t.es6_guard(x, "'__all__' assignment requires ES6")
        elements = x.value.elts
        result = [JSExport(el.s) for el in elements]
        return JSStatements(*result)

Assign = [Assign_all, Assign_default]


def AugAssign(t, x):
    return JSAugAssignStatement(x.target, x.op, x.value)


def If(t, x):
    return JSIfStatement(x.test, x.body, x.orelse)


def While(t, x):
    assert not x.orelse
    return JSWhileStatement(x.test, x.body)


def Break(t, x):
    return JSBreakStatement()


def Continue(t, x):
    return JSContinueStatement()


def Pass(t, x):
    return JSPass()


def Return(t, x):
    # x.value is None for blank return statements
    return JSReturnStatement(x.value)


def Delete(t, x):
    js = []
    for t in x.targets:
        jd = JSDeleteStatement(t)
        if len(x.targets) == 1:
            jd.py_node = x
        else:
            jd.py_node = t
        js.append(jd)
    return JSStatements(*js)


def Await(t, x):
    t.stage3_guard(x, "Async stuff requires 'stage3' to be enabled")
    return JSAwait(x.value)


#### Expressions


def Expr_default(t, x):
    # See [pj.transformations.special](special.py) for special cases
    return JSExpressionStatement(x.value)


def List(t, x):
    return JSList(x.elts)


def Tuple(t, x):
    return JSList(x.elts)


def Dict(t, x):
    return JSDict(_normalize_dict_keys(t, x.keys), x.values)


def Lambda(t, x):
    assert not any(getattr(x.args, k) for k in [
            'vararg', 'kwonlyargs', 'kwarg', 'defaults', 'kw_defaults'])
    return JSArrowFunction(
                None,
                [arg.arg for arg in x.args.args],
                [JSReturnStatement(x.body)])


def IfExp(t, x):
    return JSIfExp(x.test, x.body, x.orelse)


def Call_default(t, x, operator=None):
    # See [pj.transformations.special](special.py) for special cases
    kwkeys = []
    kwvalues = []
    if x.keywords:
        for kw in x.keywords:
            t.unsupported(x, kw.arg is None, "'**kwargs' syntax isn't "
                          "supported")
            kwkeys.append(kw.arg)
            kwvalues.append(kw.value)
        kwargs = JSDict(_normalize_dict_keys(t, kwkeys), kwvalues)
    else:
        kwargs = None
    return JSCall(x.func, x.args, kwargs, operator)


def Attribute_default(t, x):
    return JSAttribute(x.value, _normalize_name(str(x.attr)))


def Subscript_default(t, x):
    assert isinstance(x.slice, ast.Index)
    v = x.slice.value
    if isinstance(v, ast.UnaryOp) and isinstance(v.op, ast.USub):
        return JSSubscript(
            JSCall(JSAttribute(x.value, 'slice'), [v]),
            JSNum(0))
    return JSSubscript(x.value, v)


def UnaryOp(t, x):
    return JSUnaryOp(x.op, x.operand)


def BinOp_default(t, x):
    # See [pj.transformations.special](special.py) for special cases
    return JSBinOp(x.left, x.op, x.right)


def BoolOp(t, x):
    return reduce(
                lambda left, right: JSBinOp(left, x.op, right),
                x.values)


def Compare_default(t, x):
    """Compare is for those expressions like 'x in []' or 1 < x < 10. It's
    different from a binary operations because it can have multiple
    operators and more than two operands."""
    exps = [x.left] + x.comparators
    bools = []
    for i in range(len(x.ops)):
        bools.append(JSBinOp(exps[i], x.ops[i], exps[i + 1]))
    return reduce(lambda x, y: JSBinOp(x, JSOpAnd(), y), bools)


#### Atoms


def Num(t, x):
    return JSNum(x.n)


def Str(t, x):
    return JSStr(x.s)


def JoinedStr(t, x):
    t.es6_guard(x, "f-strings require ES6")
    chunks = []
    for value in x.values:
        if isinstance(value, ast.Str):
            chunks.append(value.s)
        else:
            assert isinstance(value, ast.FormattedValue)
            t.unsupported(x, value.conversion != -1,
                          "f-string conversion spec isn't supported")
            t.unsupported(x, value.format_spec is not None,
                          "f-string format spec isn't supported")
            chunks.append('${%s}' % t._transform_node(value.value))
    return JSTemplateLiteral(''.join(chunks))


def Name_default(t, x):
    # {True,False,None} are Names
    cls = {
        'True': JSTrue,
        'False': JSFalse,
        'None': JSNull,
    }.get(x.id)
    if cls:
        return cls()
    else:
        n = x.id
        n = _normalize_name(n)
        return JSName(n)


def NameConstant(t, x):
    cls = {
        True: JSTrue,
        False: JSFalse,
        None: JSNull,
    }[x.value]
    return cls()


def Yield(t, x):
    return JSYield(x.value)


def YieldFrom(t, x):
    return JSYieldStar(x.value)


#### Ops


def In(t, x):
    return JSOpIn()


def Add(t, x):
    return JSOpAdd()


def Sub(t, x):
    return JSOpSub()


def USub(t, x):
    "Handles tokens like '-1'"
    return JSOpUSub()


def Mult(t, x):
    return JSOpMult()


def Div(t, x):
    return JSOpDiv()


def Mod(t, x):
    return JSOpMod()


def RShift(t, x):
    return JSOpRShift()


def LShift(t, x):
    return JSOpLShift()


def BitXor(t, x):
    return JSOpBitXor()


def BitAnd(t, x):
    return JSOpBitAnd()


def BitOr(t, x):
    return JSOpBitOr()


def Invert(t, x):
    return JSOpInvert()


def And(t, x):
    return JSOpAnd()


def Or(t, x):
    return JSOpOr()


def Not(t, x):
    return JSOpNot()


# == and != are in special.py
# because they transform to === and !==


def Lt(t, x):
    return JSOpLt()


def LtE(t, x):
    return JSOpLtE()


def Gt(t, x):
    return JSOpGt()


def GtE(t, x):
    return JSOpGtE()


def Starred(t, x):
    return JSRest(x.value)



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- exceptions
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast


class ProcessorError(Exception):

    def __str__(self):
        py_node = self.args[0]
        if isinstance(py_node, (ast.expr, ast.stmt)):
            lineno = str(py_node.lineno)
            col_offset = str(py_node.col_offset)
        else:
            lineno = 'n. a.'
            col_offset = 'n. a.'
        return "Node type '%s': Line: %s, column: %s" % (type(py_node).__name__,
                                                         lineno,
                                                         col_offset)


class TransformationError(ProcessorError):

    def __str__(self):
        error = super().__str__()
        if len(self.args) > 1:
            error += ". %s" % self.args[1]
        return error


class UnsupportedSyntaxError(TransformationError):
    pass



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- class transformations
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast

from macropy.core.quotes import macros, ast_literal, ast_list, name, q
from macropy.experimental.pattern import (macros, _matching, switch,
    ClassMatcher, LiteralMatcher, ListMatcher)

from ..processor.util import controlled_ast_walk
from ..js_ast import (
    JSAttribute,
    JSBinOp,
    JSCall,
    JSClass,
    JSDict,
    JSExpressionStatement,
    JSList,
    JSMultipleArgsOp,
    JSName,
    JSOpInstanceof,
    JSOpOr,
    JSOpStrongEq,
    JSOpTypeof,
    JSSubscript,
    JSStatements,
    JSStr,
    JSSuper,
    JSThis,
    JSUnaryOp,
)

from .common import _build_call_isinstance

from . import _normalize_name, _normalize_dict_keys

EXC_TEMPLATE = """\
class %(name)s(Error):

    def __init__(self, message):
        self.name = '%(name)s'
        self.message = message or 'Error'
"""

EXC_TEMPLATE_ES5 = """\
def %(name)s(self, message):
    self.name = '%(name)s'
    self.message = message or 'Custom error %(name)s'
    if typeof(Error.captureStackTrace) == 'function':
        Error.captureStackTrace(self, self.constructor)
    else:
        self.stack = Error(message).stack

%(name)s.prototype = Object.create(Error.prototype)
%(name)s.prototype.constructor = %(name)s
"""


def _isdoc(el):
    return isinstance(el, ast.Expr) and isinstance(el.value, ast.Str)


def _class_guards(t, x):
    t.es6_guard(x, "'class' statement requires ES6")
    t.unsupported(x, len(x.bases) > 1, "Multiple inheritance is not supported")
    body = x.body
    for node in body:
        t.unsupported(x, not (isinstance(node, (ast.FunctionDef,
                                                ast.AsyncFunctionDef,
                                                ast.Assign)) or \
                              _isdoc(node) or isinstance(node, ast.Pass)),
                      "Class' body members must be functions or assignments")
        t.unsupported(x, isinstance(node, ast.Assign) and len(node.targets) > 1,
                      "Assignments must have only one target")
    if len(x.bases) > 0:
        assert len(x.bases) == 1
    assert not x.keywords, "class '{}', args cannot be keywords".format(x.name)


def ClassDef_exception(t, x):
    """Convert a class like::

      class MyError(Exception):
          pass

    into something like::

      class MyError extends Error {
          constructor(message) {
              this.name = 'MyError';
              this.message = message || 'Error';
          }
      }

    The real implementation avoids ES6 classes because as of now
    (2016-03-20) subclassing from Error fails the instanceof test and
    so i would break catch bodies, as for how they are transformed
    right now.

    N.B. A toString() like this is supposed to be implemented by the
    Error object:

    function toString() {
        return this.name + ': ' + this.message;
    }
    """
    # detect if the body is empty
    _class_guards(t, x)
    name = x.name
    body = x.body
    if len(x.bases) > 0 and isinstance(x.bases[0], ast.Name):
        super_name = x.bases[0].id
    else:
        super_name = None

    # strip docs from body
    fn_body = [e for e in body if isinstance(e, (ast.FunctionDef,
                                                 ast.AsyncFunctionDef))]

    # all the other kind of members which are assigned stuff
    assigns = [e for e in body if isinstance(e, ast.Assign)]

    # is this a simple definition of a subclass of Exception?
    if len(fn_body) > 0 or len(assigns) > 0 or super_name not in \
       ('Exception', 'Error'):
        return
    res = t.subtransform(EXC_TEMPLATE_ES5 % dict(name=name), remap_to=x)
    return res


def ClassDef_default(t, x):
    """Convert a class to an ES6 class."""

    # check if translatable
    _class_guards(t, x)

    name = x.name
    body = x.body

    if len(x.bases) > 0:
        superclass = x.bases[0]
    else:
        superclass = None

    # strip docs from body
    fn_body = [e for e in body if isinstance(e, (ast.FunctionDef,
                                                 ast.AsyncFunctionDef))]

    # all the other kind of members which are assigned stuff
    assigns = [e for e in body if isinstance(e, ast.Assign)]

    # Each FunctionDef must have self as its first arg
    # silly check for methods
    for node in fn_body:
        arg_names = [arg.arg for arg in node.args.args]
        t.unsupported(node, len(arg_names) == 0 or arg_names[0] != 'self',
                      "First arg on method must be 'self'")

    # TODO: better express this... find if the constructor has to be the first
    # as per ES6 doc
    if len(fn_body) > 0 and fn_body[0].name == '__init__':
        init = body[0]
        # __init__ should not contain a return statement
        # silly check
        for stmt in controlled_ast_walk(init):
            assert not isinstance(stmt, ast.Return)

    # manage decorators. some are managed at class translation time and
    # converted to equal ES6 syntax while the generic ones will be calculated
    # at runtime
    decos = {}
    for fn in fn_body:
        # make sure the function hasn't any decorator managed by the function
        # transformer. This includes property, property, classmethod,
        # staticmethod and property.setter:
        if fn.decorator_list and not \
           ((len(fn.decorator_list) == 1 and
            isinstance(fn.decorator_list[0], ast.Name) and
            fn.decorator_list[0].id in ['property', 'classmethod',
                                        'staticmethod']) or
            (isinstance(fn.decorator_list[0], ast.Attribute) and
             fn.decorator_list[0].attr == 'setter')):

            decos[fn.name] = (JSStr(fn.name), fn.decorator_list)
            fn.decorator_list = []  # remove so that the function transformer
            # will not complain

    # keep class doc if present
    if _isdoc(body[0]):
        fn_body = [body[0]] + fn_body

    # assign incoming pynode to the JSClass for the sourcemap
    cls = JSClass(JSName(name), superclass, fn_body)
    cls.py_node = x

    stmts = [cls]

    # prepare assignments mapping as js ast
    def _from_assign_to_dict_item(e):
        key = e.targets[0]
        value = e.value
        if isinstance(key, ast.Name):
            rendered_key = ast.Str(_normalize_name(key.id))
            sort_key = key.id
        else:
            rendered_key = key
            sort_key = '~'
        return sort_key, rendered_key, value

    assigns = tuple(zip(*sorted(map(_from_assign_to_dict_item, assigns),
                                key=lambda e: e[0])))

    # render assignments as properties at runtime
    if assigns:
        from ..snippets import set_properties
        t.add_snippet(set_properties)
        assigns = JSExpressionStatement(
            JSCall(
                JSAttribute(JSName('_pj'), 'set_properties'),
                (JSName(name),
                 JSDict(_normalize_dict_keys(t, assigns[1]), assigns[2])),
            )
        )
        stmts.append(assigns)

    # calculate method decorators at runtime
    if decos:
        from ..snippets import set_decorators
        t.add_snippet(set_decorators)
        keys = []
        values = []
        for k, v in sorted(decos.items(), key=lambda i: i[0]):
            rendered_key, dlist = v
            keys.append(rendered_key)
            values.append(JSList(dlist))
        decos = JSExpressionStatement(
            JSCall(
                JSAttribute(JSName('_pj'), 'set_decorators'),
                (JSName(name),
                 JSDict(_normalize_dict_keys(t, keys), values)),
            )
        )
        stmts.append(decos)

    # there is any decorator list on the class
    if x.decorator_list:
        from ..snippets import set_class_decorators
        t.add_snippet(set_class_decorators)
        with q as cls_decos:
            name[name] = _pj.set_class_decorators(
                name[name], ast_list[x.decorator_list])

        stmts.append(JSExpressionStatement(cls_decos[0]))
    return JSStatements(*stmts)


ClassDef = [ClassDef_exception, ClassDef_default]


def Call_super(t, x):
    if isinstance(x.func, ast.Attribute) and isinstance(x.func.value, ast.Call) \
         and isinstance(x.func.value.func, ast.Name) and \
         x.func.value.func.id == 'super':
        sup_args = x.func.value.args
        # Are we in a FuncDef and is it a method and super() has no args?
        method = t.find_parent(x, ast.FunctionDef, ast.AsyncFunctionDef)
        if method and isinstance(t.parent_of(method), ast.ClassDef) and \
           len(sup_args) == 0:
            # if in class constructor, this becomes ``super(x, y)``
            if method.name == '__init__':
                result = JSCall(JSSuper(), x.args)
            else:
                sup_method = x.func.attr
                if isinstance(method, ast.AsyncFunctionDef):
                    # temporary fix for babel's bug
                    # https://github.com/babel/babel/issues/3930
                    # converts to
                    # O.getPrototypeOf(O.getPrototypeOf(this)).method.call(this,
                    # x, y)
                    sup_cls = t.parent_of(method).bases[0]
                    result = JSCall(
                        JSAttribute(
                            JSAttribute(
                                JSAttribute(JSName(sup_cls),
                                                   'prototype'),
                                _normalize_name(sup_method)),
                            'call'),
                        [JSThis()] + x.args)
                else:
                # this becomes super.method(x, y)
                    result = JSCall(
                        JSAttribute(JSSuper(), _normalize_name(sup_method)),
                        x.args
                    )
            return result


def Attribute_super(t, x):
    """Translate ``super().foo`` into ``super.foo` if the method isn't a constructor,
    where it's invalid.

    AST is::

      Attribute(attr='foo',
                ctx=Load(),
                value=Call(args=[],
                           func=Name(ctx=Load(),
                                     id='super'),
                           keywords=[]))

    """
    if isinstance(x.value, ast.Call) and len(x.value.args) == 0 and \
       isinstance(x.value.func, ast.Name) and x.value.func.id == 'super':
        sup_args = x.value.args
        # Are we in a FuncDef and is it a method and super() has no args?
        method = t.find_parent(x, ast.FunctionDef, ast.AsyncFunctionDef)
        if method and isinstance(t.parent_of(method), ast.ClassDef) and \
           len(sup_args) == 0:
            if method.name == '__init__':
                t.unsupported(x, True, "'super().attr' cannot be used in "
                              "constructors")
            else:
                sup_method = x.attr
                # this becomes super.method
                result = JSAttribute(JSSuper(), _normalize_name(sup_method))
            return result


def Subscript_super(t, x):
    """Same as per attribute: translate ``super()[foo]`` into ``super[foo]``.

    AST is::

         Subscript(ctx=Load(),
                   slice=Index(value=Name(ctx=Load(),
                                          id='foo')),
                   value=Call(args=[],
                              func=Name(ctx=Load(),
                                        id='super'),
                              keywords=[]))

    """
    if isinstance(x.value, ast.Call) and isinstance(x.value.func, ast.Name) and \
       x.value.func.id == 'super':
        sup_args = x.value.args
        # Are we in a FuncDef and is it a method and super() has no args?
        method = t.find_parent(x, ast.FunctionDef, ast.AsyncFunctionDef)
        if method and isinstance(t.parent_of(method), ast.ClassDef) and \
           len(sup_args) == 0:
            if method.name == '__init__':
                t.unsupported(x, True, "'super()[expr]' cannot be used in "
                              "constructors")
            else:
                sup_method = x.slice.value
                # this becomes super[expr]
                result = JSSubscript(JSSuper(), _normalize_name(sup_method))
            return result


def Call_isinstance(t, x):
    """Translate ``isinstance(foo, Bar)`` to ``foo instanceof Bar`` and
    ``isinstance(Foo, (Bar, Zoo))`` to ``foo instanceof Bar || foo instanceof
    Zoo``.

    AST dump of the latter::

      Call(args=[Name(ctx=Load(),
                      id='foo'),
                 Tuple(ctx=Load(),
                       elts=[Name(ctx=Load(),
                                  id='Bar'),
                             Name(ctx=Load(),
                                  id='Zoo')])],
           func=Name(ctx=Load(),
                     id='isinstance'),
           keywords=[])

    """
    if (isinstance(x.func, ast.Name) and x.func.id == 'isinstance'):
        assert len(x.args) == 2
        return _build_call_isinstance(x.args[0], x.args[1])


def Call_issubclass(t, x):
    """Translate ``issubclass(Foo, Bar)`` to ``Foo.prototype instanceof Bar``.
    """
    with switch(x):
        if ast.Call(func=ast.Name(id='issubclass'), args=[target, classes]):
            tproto = q[ast_literal[target].prototype]
            if isinstance(classes, (ast.Tuple, ast.List, ast.Set)):
                classes = classes.elts
            else:
                classes = [classes]
            prev = None
            for c in classes:
                cur = q[ast_literal[c].prototype.isPrototypeOf(
                    ast_literal[tproto])]
                if prev is not None:
                    cur = q[ast_literal[prev] or ast_literal[cur]]
                prev = cur
            return JSExpressionStatement(cur)



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj
# :Created:  gio 09 mar 2017 19:43:11 CET
# :Author:   Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
import logging

import macropy.activate

from ..js_ast import JSKeySubscript, JSStr, TargetNode

logger = logging.getLogger(__name__)


def _normalize_name(n):
    if n.startswith('d_'):
        n = n.replace('d_', '$')
    elif n.startswith('dd_'):
        n = n.replace('dd_', '$$')
    elif not n.startswith('_') and n.endswith('_'):
        n = n[:-1]
    return n

def _normalize_dict_keys(transformer, keys):
    res = []
    for key in keys:
        if isinstance(key, str):
            key = ast.Str(key)
        elif isinstance(key, JSStr):
            key = ast.Str(key.args[0])
        if not isinstance(key, ast.Str):
            if transformer.enable_es6:
                key = JSKeySubscript(key)
            else:
                if isinstance(key, ast.AST):
                    py_node = key
                elif isinstance(key, TargetNode) and key.py_node is not None:
                    py_node = key.py_node
                else:
                    raise ValueError('Value of type %r cannot '
                                     'be use as key' % type(key))
                transformer.unsupported(py_node, True, 'Value of type %r cannot '
                                        'be use as key' % type(key))
        res.append(key)
    return res



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- exceptions transformations
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast

from ..js_ast import (
    JSBinOp,
    JSIfStatement,
    JSName,
    JSNewCall,
    JSOpInstanceof,
    JSStatements,
    JSThrowStatement,
    JSTryCatchFinallyStatement,
    JSVarStatement,
)

from .common import _build_call_isinstance


def Try(t, x):
    t.unsupported(x, x.orelse, "'else' block of 'try' statement isn't supported")
    known_exc_types = (ast.Name, ast.Attribute, ast.Tuple, ast.List)
    ename = None
    if x.handlers:
        for h in x.handlers:
            if h.type is not None and not isinstance(h.type, known_exc_types):
                t.warn(x, "Exception type expression might not evaluate to a "
                       "valid type or sequence of types.")
            ename = h.name
        ename = ename or 'e'
        if t.has_child(x.handlers, ast.Raise) and t.has_child(x.finalbody, ast.Return):
            t.warn(x, node, "The re-raise in 'except' body may be masked by the "
                   "return in 'final' body.")
            # see
            # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Control_flow_and_error_handling#The_finally_block

        rhandlers = x.handlers.copy()
        rhandlers.reverse()
        prev_except = stmt =  None
        for ix, h in enumerate(rhandlers):
            body = h.body
            if h.name is not None and h.name != ename:
                # Rename the exception to match the handler
                rename = JSVarStatement([h.name], [ename])
                body = [rename] + h.body

            # if it's  the last except and it's a catchall
            # threat 'except Exception:' as a catchall
            if (ix == 0 and h.type is None or (isinstance(h.type, ast.Name) and
                                   h.type.id == 'Exception')):
                prev_except = JSStatements(*body)
                continue
            else:
                if ix == 0:
                    prev_except = JSThrowStatement(JSName(ename))
                # then h.type is an ast.Name != 'Exception'
                stmt = JSIfStatement(
                    _build_call_isinstance(JSName(ename), h.type),
                    body,
                    prev_except
                )
            prev_except = stmt
        t.ctx['ename'] = ename
        result = JSTryCatchFinallyStatement(x.body, ename, prev_except, x.finalbody)
    else:
        result = JSTryCatchFinallyStatement(x.body, None, None, x.finalbody)
    return result


def Raise(t, x):
    if x.exc is None:
        ename = t.ctx.get('ename')
        t.unsupported(x, not ename, "'raise' has no argument but failed obtaining"
                      " implicit exception")
        res = JSThrowStatement(JSName(ename))
    elif isinstance(x.exc, ast.Call) and isinstance(x.exc.func, ast.Name) and \
         len(x.exc.args) == 1 and x.exc.func.id == 'Exception':
        res = JSThrowStatement(JSNewCall(JSName('Error'), x.exc.args))
    else:
        res = JSThrowStatement(x.exc)

    return res



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- public api
# :Created:  ven 26 feb 2016 15:16:13 CET
# :Author:   Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
import base64
import json
import inspect
import logging
import os
import textwrap

import dukpy

from .processor.transforming import Transformer
from .processor.util import Block
from .js_ast import JSStatements
from . import transformations

log = logging.getLogger(__name__)

BABEL_COMPILER = os.path.join(os.path.dirname(__file__), 'data',
                              'babel-6.18.1.min.js')
BABEL_POLYFILL = os.path.join(os.path.dirname(__file__), 'data',
                              'polyfill.min.js')


def _calc_file_names(src_filename, dst_filename=None, map_filename=None):
    """Calculate destination paths for file translation/transpile."""
    src_filename = os.path.abspath(src_filename)
    src_dir = os.path.dirname(src_filename)
    if dst_filename and not os.path.isdir(dst_filename):
        dst_filename = os.path.abspath(dst_filename)
        dst_dir = os.path.dirname(dst_filename)
    else:
        if dst_filename and os.path.isdir(dst_filename):
            dst_dir = os.path.abspath(dst_filename)
        else:
            dst_dir = src_dir
        dst_name = os.path.basename(os.path.splitext(src_filename)[0])
        dst_filename = os.path.join(dst_dir, dst_name + '.js')
    if map_filename:
        map_filename = os.path.abspath(map_filename)
    else:
        map_filename = dst_filename + '.map'
    map_dir = os.path.dirname(map_filename)
    src_relpath = os.path.relpath(src_filename, map_dir)
    map_relpath = os.path.relpath(map_filename, dst_dir)
    return dst_filename, map_filename, src_relpath, map_relpath


def _inline_src_map(src_map):
    src_map_data = ('data:text/json;base64,%s' %
        base64.b64encode(src_map.encode('utf-8')).decode('ascii'))
    return '\n//# sourceMappingURL=%s\n' % src_map_data


def translate_file(src_filename, dst_filename=None, map_filename=None,
                   enable_es6=False, enable_stage3=False, inline_map=False):
    """Translate the given python source file to ES6 Javascript."""
    dst_filename, map_filename, src_relpath, map_relpath = _calc_file_names(
        src_filename, dst_filename, map_filename
    )
    src_text = open(src_filename).readlines()
    js_text, src_map = translates(src_text, True, src_relpath,
                                  enable_es6=enable_es6,
                                  enable_stage3=enable_stage3)
    if inline_map:
        js_text += _inline_src_map(src_map)
    else:
        js_text += '\n//# sourceMappingURL=%s\n' % map_relpath

    with open(dst_filename, 'w') as dst:
        dst.write(js_text)
    if not inline_map:
        with open(map_filename, 'w') as map:
            map.write(src_map)


def translate_object(py_obj, body_only=False, enable_es6=False,
                     enable_stage3=False):
    """Translate the given Python 3 object (function, class, etc.) to ES6
    Javascript.

    If `body_only` is ``True``, the object itself is discarded and only its
    body gets translated as it was a module body.

    Return a ``(js_text, js_source_map)`` tuple.
    """
    cwd = os.getcwd()
    src_filename = os.path.abspath(inspect.getsourcefile(py_obj))
    prefix = os.path.commonpath((cwd, src_filename))
    if len(prefix) > 1:
        src_filename = src_filename[len(prefix) + 1:]
    src_lines, sline_offset = inspect.getsourcelines(py_obj)
    # line offsets should be 0-based
    sline_offset = sline_offset - 1
    with open(src_filename) as f:
        complete_src = f.read()
    return translates(src_lines, True, src_filename, (sline_offset, 0),
                      body_only=body_only, complete_src=complete_src,
                      enable_es6=enable_es6, enable_stage3=enable_stage3)


def translates(src_text, dedent=True, src_filename=None, src_offset=None,
               body_only=False, complete_src=None, enable_es6=False,
               enable_stage3=False):
    """Translate the given Python 3 source text to ES6 Javascript.

    If the string comes from a file, it's possible to specify the filename
    that will be inserted into the output source map. The `src_offset` is the
    ``(line_offset, col_offset)`` tuple of the fragment and it's used to
    relocate the map segments. `map_filename` is the intended file name for
    the output map file that will be added as pragma comment to the output JS.

    Setting `body_only` to a true value will change the evaluation behavior to
    translate only the body of the first statement.
    """
    if isinstance(src_text, (tuple, list)):
        src_lines = src_text
        src_text = ''.join(src_text)
    else:
        src_lines = src_text.splitlines() # removes \n

    # take into account only the lines with content because only those
    # will be dedented
    src_line_num = 0
    for l in src_lines:
        if not len(l.strip()) == 0:
            src_line_num += 1

    sline_offset, scol_offset = src_offset or (0, 0)
    if dedent:
        # remove indentation so that fragments of files with stuff not
        # at root can be evaluated, or ast will complain
        dedented = textwrap.dedent(src_text)
        if len(dedented) < len(src_text):
            scol_offset += (len(src_text) - len(dedented)) // src_line_num
    else:
        dedented = src_text
    t = Transformer(transformations, JSStatements, es6=enable_es6,
                    stage3=enable_stage3)
    pyast = ast.parse(dedented)
    if body_only and hasattr(pyast, 'body') and len(pyast.body) == 1 \
       and hasattr(pyast.body[0], 'body'):
        # if body_only is true, discard the top level element and and
        # take the first child as top, this will allow to use the body
        # of a function or class as it was a module bosy
        pyast = pyast.body[0]
        # naive check, remove the last statement if it's a return
        if isinstance(pyast.body[-1], ast.Return):
            pyast.body.pop()
    jsast = t.transform_code(pyast)
    dline_offset = dcol_offset = 0
    if t.snippets:
        snipast = t.transform_snippets()
        snipast += jsast
        jsast = snipast
    js_code_block = Block(jsast)
    js_text = js_code_block.read()
    enc_src_map, src_map = js_code_block.sourcemap(complete_src or src_text,
                                                   src_filename,
                                                   (sline_offset, scol_offset),
                                                   (dline_offset, dcol_offset))
    for t in src_map.tokens:
        log.debug("js: (%d, %d)\t\t py: (%d, %d)\t txt: '%s'",
                  t.dst_line, t.dst_col, t.src_line - sline_offset,
                  t.src_col - scol_offset, t.mapping['text'])
    return js_text, enc_src_map


def transpile_es6s(es6_text, es6_filename=None, es6_sourcemap=None,
                   enable_stage3=False, **kw):
    """Transpile the given ES6 Javascript to ES5 Javascript using Dukpy
    and babeljs."""
    opts = dict(sourceMaps=True)
    if es6_filename:
        opts['filename'] = es6_filename
    if es6_sourcemap:
        if isinstance(es6_sourcemap, str):
            es6_sourcemap = json.loads(es6_sourcemap)
        opts['inputSourceMap'] = es6_sourcemap
    opts['presets'] = ["es2015"]
    if enable_stage3:
        opts['presets'].append('stage-3')
    truntime = kw.pop('truntime', False)
    if truntime:
        opts['plugins'] = ['transform-runtime']
    res = babel_compile(es6_text, **opts)
    return res['code'], res['map']


def transpile_object(py_obj, body_only=False, es6_filename=None,
                     enable_stage3=False, **kw):
    """Transpile the given Python 3 object to ES5 Javascript using
    Dukpy and babeljs."""
    es6_text, es6_sourcemap = translate_object(py_obj, body_only=body_only,
                                               enable_es6=True,
                                               enable_stage3=enable_stage3)
    return transpile_es6s(es6_text, es6_filename, es6_sourcemap,
                          enable_stage3=enable_stage3, **kw)


def transpile_pys(src_text, dedent=True, src_filename=None, src_offset=None,
                  body_only=False, es6_filename=None, enable_stage3=False, **kw):
    """Transpile the given Python 3 source text to ES5 Javascript
    using Dukpy and babeljs.
    """
    es6_text, es6_sourcemap = translates(src_text, dedent, src_filename,
                                         src_offset, body_only, enable_es6=True,
                                         enable_stage3=enable_stage3)
    return transpile_es6s(es6_text, es6_filename, es6_sourcemap,
                          enable_stage3=enable_stage3, **kw)


def transpile_py_file(src_filename, dst_filename=None, map_filename=None,
                      enable_stage3=False, **kw):
    """Transpile the given Python 3 source file to ES5 Javascript
    using Dukpy and babeljs.
    """
    dst_filename, map_filename, src_relpath, map_relpath = _calc_file_names(
        src_filename, dst_filename, map_filename
    )
    dst_name, dst_ext = os.path.splitext(dst_filename)
    map_name, map_ext = os.path.splitext(map_filename)
    es6_dst_filename = dst_name + '.es6' + dst_ext
    es6_map_filename = map_name + '.es6' + map_ext
    dst_dir = os.path.dirname(dst_filename)
    es6_map_relpath = os.path.relpath(es6_map_filename, dst_dir)
    es6_relpath = os.path.relpath(es6_dst_filename, dst_dir)
    src_text = open(src_filename).readlines()
    es6_text, es6_src_map = translates(src_text, True, src_relpath,
                                       enable_es6=True, enable_stage3=enable_stage3)

    es5_text, es5_src_map = transpile_es6s(es6_text, es6_relpath, es6_src_map,
                                           enable_stage3=enable_stage3, **kw)
    es5_text += '\n//# sourceMappingURL=%s\n' % map_relpath
    es6_text += '\n//# sourceMappingURL=%s\n' % es6_map_relpath

    with open(es6_dst_filename, 'w') as dst:
        dst.write(es6_text)
    with open(es6_map_filename, 'w') as map:
        map.write(es6_src_map)

    with open(dst_filename, 'w') as dst:
        dst.write(es5_text)
    with open(map_filename, 'w') as map:
        map.write(json.dumps(es5_src_map))


def evaljs(js_text, load_es6_polyfill=False, **kwargs):
    """Evaluate JS code, like ``dukpy.evaljs()``, optionally loading the es6
    polyfill before the actual code.
    """
    if isinstance(js_text, (str, bytes, bytearray)):
        js_text = [js_text]
    else:
        list(js_text)
    if load_es6_polyfill:
        with open(BABEL_POLYFILL, 'r') as babel_poly:
            js_text = ['global = this; this.console = {log: print};\n'] + [babel_poly.read()] + js_text
    return dukpy.evaljs(js_text, **kwargs)


def eval_object(py_obj, append=None, body_only=False, ret_code=False,
                **kwargs):
    js_text, _ = translate_object(py_obj, body_only)
    if append:
        js_text += append
    res = dukpy.evaljs(js_text, **kwargs)
    if ret_code:
        res = (res, js_text)
    return res


def evals(py_text, body_only=False, ret_code=False, **kwargs):
    js_text, _ = translates(py_text, body_only=body_only)
    res = dukpy.evaljs(js_text, **kwargs)
    if ret_code:
        res = (res, js_text)
    return res


def eval_object_es6(py_obj, append=None, body_only=False, ret_code=False,
                    enable_stage3=False, **kwargs):
    es5_text, _ = transpile_object(py_obj, body_only, enable_stage3=enable_stage3)
    if append:
        es5_text += '\n' + append
    res = evaljs(es5_text, load_es6_polyfill=True, **kwargs)
    if ret_code:
        res = (res, es5_text)
    return res


def evals_es6(py_text, body_only=False, ret_code=False, enable_stage3=False,
              **kwargs):
    es5_text, _ = transpile_pys(py_text, body_only=body_only,
                                enable_stage3=enable_stage3)
    res = evaljs(es5_text, load_es6_polyfill=True, **kwargs)
    if ret_code:
        res = (res, es5_text)
    return res

BABEL_JS_CTX = None

def babel_compile(source, reuse_js_ctx=True, **kwargs):
    """Compile the given `source` from ES6 to ES5 usin Babeljs."""
    global BABEL_JS_CTX
    presets = kwargs.get('presets')
    if not presets:
        kwargs['presets'] = ["es2015"]
    trans_code = ('var bres, res;'
                  'bres = Babel.transform(dukpy.es6code, dukpy.babel_options);',
                  'res = {map: bres.map, code: bres.code};')
    if reuse_js_ctx and BABEL_JS_CTX:
        result = BABEL_JS_CTX.evaljs(trans_code, es6code=source,
                                     babel_options=kwargs)
    else:
        with open(BABEL_COMPILER, 'r') as babel_js:
            if reuse_js_ctx:
                BABEL_JS_CTX = dukpy.JSInterpreter()
                eval_fn = BABEL_JS_CTX.evaljs
            else:
                eval_fn = dukpy.evaljs
            result = eval_fn(
                (babel_js.read(),
                 'var bres, res;'
                 'bres = Babel.transform(dukpy.es6code, dukpy.babel_options);',
                 'res = {map: bres.map, code: bres.code};'),
                es6code=source,
                babel_options=kwargs
            )
    return result



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- comprehensions transformations
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast

from ..js_ast import (
    JSAttribute,
    JSAugAssignStatement,
    JSBinOp,
    JSCall,
    JSExpressionStatement,
    JSForStatement,
    JSFunction,
    JSIfStatement,
    JSList,
    JSName,
    JSNum,
    JSOpAdd,
    JSOpLt,
    JSReturnStatement,
    JSSubscript,
    JSThis,
    JSVarStatement,
)

#### ListComp
# Transform
# <pre>[EXPR for NAME in LIST]</pre>
# or
# <pre>[EXPR for NAME in LIST if CONDITION]</pre>
def ListComp(t, x):

    assert len(x.generators) == 1
    assert len(x.generators[0].ifs) <= 1
    assert isinstance(x.generators[0], ast.comprehension)
    assert isinstance(x.generators[0].target, ast.Name)

    EXPR = x.elt
    NAME = x.generators[0].target
    LIST = x.generators[0].iter
    if len(x.generators[0].ifs) == 1:
        CONDITION = x.generators[0].ifs[0]
    else:
        CONDITION = None

    __new = t.new_name()
    __old = t.new_name()
    __i = t.new_name()
    __bound = t.new_name()

    # Let's contruct the result from the inside out:
    #<pre>__new.push(EXPR);</pre>
    push = JSExpressionStatement(
            JSCall(
                JSAttribute(
                    JSName(__new),
                    'push'),
                [EXPR]))

    # If needed, we'll wrap that with:
    #<pre>if (CONDITION) {
    #    <i>...push...</i>
    #}</pre>
    if CONDITION:
        pushIfNeeded = JSIfStatement(
                CONDITION,
                push,
                None)
    else:
        pushIfNeeded = push

    # Wrap with:
    #<pre>for(
    #        var __i = 0, __bound = __old.length;
    #        __i &lt; __bound;
    #        __i++) {
    #    var NAME = __old[__i];
    #    <i>...pushIfNeeded...</i>
    #}</pre>
    forloop = JSForStatement(
                    JSVarStatement(
                        [__i, __bound],
                        [0, JSAttribute(
                                JSName(__old),
                                'length')]),
                    JSBinOp(JSName(__i), JSOpLt(), JSName(__bound)),
                    JSAugAssignStatement(JSName(__i), JSOpAdd(), JSNum(1)),
                    [
                        JSVarStatement(
                            [NAME.id],
                            [JSSubscript(
                                JSName(__old),
                                JSName(__i))]),
                        pushIfNeeded])

    # Wrap with:
    #<pre>function() {
    #    var __new = [], __old = LIST;
    #    <i>...forloop...</i>
    #    return __new;
    #}
    func = JSFunction(
        None,
        [],
        [
            JSVarStatement(
                [__new, __old],
                [JSList([]), LIST]),
            forloop,
            JSReturnStatement(
                JSName(__new))])

    # And finally:
    #<pre>((<i>...func...</i>).call(this))</pre>
    invoked = JSCall(
            JSAttribute(
                func,
                'call'),
            [JSThis()])

    return invoked



# -*- coding: utf-8 -*-
# Copyright (c) Waldemar Kornewald
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of the copyright holder nor the names of its
#        contributors may be used to endorse or promote products
#        derived from this software without specific prior written
#        permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from bisect import bisect_right
from collections import namedtuple
import json
import re
import sys

PY3 = sys.version_info[0] == 3
text_type = str if PY3 else unicode

# A single base 64 digit can contain 6 bits of data. For the base 64
# variable length quantities we use in the source map spec, the first
# bit is the sign, the next four bits are the actual value, and the
# 6th bit is the continuation bit. The continuation bit tells us
# whether there are more digits in this value following this digit.
#
#   Continuation
#   |    Sign
#   |    |
#   V    V
#   101011
VLQ_BASE_SHIFT = 5

# binary: 100000
VLQ_BASE = 1 << VLQ_BASE_SHIFT

# binary: 011111
VLQ_BASE_MASK = VLQ_BASE - 1

# binary: 100000
VLQ_CONTINUATION_BIT = VLQ_BASE

BASE64_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
BASE64_CHAR_TO_INT = dict(map(reversed, enumerate(BASE64_CHARS)))


class SourceMapDecodeError(ValueError):
    pass


def decode_vlqs(segment):
    """
    Decode a string of VLQ-encoded data.

    Returns:
      a list of integers.
    """

    values = []

    vlq = shift = 0
    for c in segment:
        val = BASE64_CHAR_TO_INT[c]
        cont = val & VLQ_CONTINUATION_BIT
        # Each character is 6 bits:
        # 5 of value and the high bit is the continuation.
        val &= VLQ_BASE_MASK
        vlq += val << shift
        shift += VLQ_BASE_SHIFT

        if not cont:
            # The low bit of the unpacked value is the sign.
            vlq, sign = vlq >> 1, vlq & 1
            if sign:
                vlq = -vlq
            values.append(vlq)
            vlq = shift = 0

    if vlq or shift:
        raise SourceMapDecodeError('leftover vlq/shift in vlq decode')

    return values


def encode_vlq(num):
    """Encode a single number in VLQ format"""
    vlq = (-num << 1) + 1 if num < 0 else num << 1
    result = ''
    while not result or vlq > 0:
        digit = vlq & VLQ_BASE_MASK
        vlq >>= VLQ_BASE_SHIFT
        if vlq > 0:
            digit |= VLQ_CONTINUATION_BIT
        result += BASE64_CHARS[digit]
    return result


def decode(source, ignore_errors=True):
    if isinstance(source, dict):
        smap = source
    else:
        # According to the spec a souce map may be prepended with
        # ")]}'" to cause a JavaScript error. In that case ignore the
        # entire first line.
        if source[:4] == ")]}'":
            source = source.split('\n', 1)[1]
        smap = json.loads(source)

    sources = smap['sources']
    names = list(map(text_type, smap['names']))
    lines = smap['mappings'].split(';')

    source_root = smap.get('source_root')
    if source_root is not None:
        sources = ['/'.join((source_root, s)) for s in sources]

    tokens = []

    dst_col = src_id = src_line = src_col = name_id = 0
    for dst_line, line in enumerate(lines):
        dst_col = 0
        for segment in line.split(','):
            if not segment:
                continue
            fields = decode_vlqs(segment)
            dst_col += fields[0]
            if dst_col < 0:
                raise SourceMapDecodeError(
                    'Segment {} has negative dst_col'.format(segment, fields))

            src = None
            name = None
            if len(fields) not in (1, 4, 5):
                raise SourceMapDecodeError(
                    'Invalid segment {}, parsed as {}'.format(segment, fields))


            if len(fields) > 1:
                src_id += fields[1]
                if not 0 <= src_id < len(sources):
                    raise SourceMapDecodeError(
                        'Segment {} references source {} which does not exist'.format(
                        segment, src_id))
                src = sources[src_id]
                src_line += fields[2]
                if src_line < 0:
                    raise SourceMapDecodeError('Segment {} has negative src_line'.format(
                                               segment))
                src_col += fields[3]
                if src_col < 0:
                    raise SourceMapDecodeError('Segment {} has negative src_col'.format(
                                               segment))

            if len(fields) > 4:
                name_id += fields[4]
                if not 0 <= name_id < len(names):
                    raise SourceMapDecodeError(
                        'Segment {} references name {} which does not exist'.format(
                        segment, name_id))
                name = names[name_id]

            tokens.append(Token(dst_line, dst_col, src, src_line, src_col, name))

    sources_content = {src: content
                       for src, content in zip(sources, smap.get('sourcesContent',
                                                                 (None,) * len(sources)))
                       if content is not None}
    return SourceMap(tokens, sources_content, raw=smap,
                     ignore_errors=ignore_errors)


def encode(sourcemap):
    sources = {}
    prev_src_id = next_src_id = 0
    prev_src_line = prev_src_col = 0
    names = {}
    prev_name_id = next_name_id = 0
    mappings = []
    prev_dst_line = -1
    for token in sourcemap.tokens:
        while prev_dst_line < token.dst_line:
            prev_dst_line += 1
            prev_dst_col = 0
            segments = []
            mappings.append(segments)
        vlq = [token.dst_col - prev_dst_col]
        prev_dst_col = token.dst_col
        if token.src:
            source_id = sources.get(token.src)
            if source_id is None:
                sources[token.src] = source_id = next_src_id
                next_src_id += 1
            vlq.append(source_id - prev_src_id)
            vlq.append(token.src_line - prev_src_line)
            vlq.append(token.src_col - prev_src_col)
            if token.name:
                name_id = names.get(token.name)
                if name_id is None:
                    names[token.name] = name_id = next_name_id
                    next_name_id += 1
                vlq.append(name_id - prev_name_id)
                prev_name_id = name_id
            prev_src_id = source_id
            prev_src_line = token.src_line
            prev_src_col = token.src_col
        segments.append(''.join(map(encode_vlq, vlq)))
    data = {'version': 3,
            'mappings': ';'.join(map(','.join, mappings)),
            'sources': sorted(sources, key=lambda x: sources[x]),
            'names': sorted(names, key=lambda x: names[x])}
    data['sourcesContent'] = list(map(sourcemap.sources_content.get, data['sources']))
    return json.dumps(data)

source_map_url_re = re.compile(r'/[\*/][#@]\s*sourceMappingURL=([^\s*]+)\s*(?:\*/)?')


def discover(source):
    source = source.splitlines()
    # Source maps are only going to exist at either the top or bottom
    # of the document.  Technically, there isn't anything indicating
    # *where* it should exist, so we are generous and assume it's
    # somewhere either in the first or last 5 lines.  If it's
    # somewhere else in the document, you're probably doing it wrong.
    if len(source) > 10:
        possibilities = source[:5] + source[-5:]
    else:
        possibilities = source

    for line in set(possibilities):
        result = source_map_url_re.findall(line)
        if result:
            return result[0]
    return None

def strip(content):
    return source_map_url_re.sub('', content)

# namedtuples have a nice repr and they support comparison (useful for
# bisect search)
class Token(namedtuple('TokenBase', 'dst_line dst_col src src_line src_col name'
                       ' mapping')):
    __slots__ = ()
    def __new__(cls, dst_line=0, dst_col=0, src='', src_line=0, src_col=0,
                name=None, mapping=None):
        return super(Token, cls).__new__(cls, dst_line, dst_col,
                                         src, src_line, src_col, name, mapping)

def shift_tokens(tokens, dst_line=0, dst_col=0, src_line=0, src_col=0):
    return [t._replace(dst_line=t.dst_line + dst_line,
                       dst_col=t.dst_col + dst_col,
                       src_line=t.src_line + src_line,
                       src_col=t.src_col + src_col)
            for t in tokens]

TOKEN_SPEC = (
    ('LINECOMMENT', r'//'),
    ('COMMENTSTART', r'/\*'),
    ('COMMENTEND', r'\*/'),
    ('DELIM',   r'[{}\[\]();]'),
)
token_re = re.compile('|'.join('(?P<%s>%s)' % pair for pair in TOKEN_SPEC),
                      re.UNICODE | re.MULTILINE | re.DOTALL)

def identity_tokenize(content, src):
    in_comment = False
    for line_num, line in enumerate(content.splitlines()):
        if line.strip():
            start_column = len(line) - len(line.lstrip())
            if start_column:
                yield Token(line_num, 0, src, line_num, 0)
            yield Token(line_num, start_column, src, line_num, start_column)
        for match in token_re.finditer(line):
            kind = match.lastgroup
            value = match.group(kind)
            if in_comment:
                if kind == 'COMMENTEND' and in_comment:
                    in_comment = False
            elif kind == 'COMMENTSTART':
                in_comment = True
            elif kind == 'LINECOMMENT':
                break
            elif kind == 'DELIM':
                column = match.start()
                if column > start_column:
                    yield Token(line_num, column, src, line_num, column)

def identity_map(content, src):
    return SourceMap(identity_tokenize(content, src), {src: content})

class SourceMap(object):
    def __init__(self, tokens=(), sources_content=None, raw=None, ignore_errors=False):
        self.tokens = []
        self.sources_content = {} if sources_content is None else sources_content.copy()
        self.raw = {} if raw is None else raw.copy()
        self.ignore_errors = ignore_errors
        map(self.add_token, tokens)

    def add_token(self, token):
        if not self.tokens or token[:2] > self.tokens[-1][:2]:
            self.tokens.append(token)
        else:
            index = bisect_right(self.tokens, token)
            if not self.ignore_errors and index and \
                    self.tokens[index - 1][:2] == token[:2]:
                raise ValueError('Token with given dst_line and dst_col already exists:\n'
                                 'Existing: {}\nAdded: {}'.format(self.tokens[index - 1],
                                                                  token))
            self.tokens.insert(index, token)



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- transformation utilities
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
import inspect
import re
import textwrap

from . import sourcemaps


IGNORED_NAMES = ('__all__',)


def delimited(delimiter, arr, dest=None, at_end=False):
    """Similar to ``str.join()``, but returns an array with an option to append the
    delimiter at the end.
    """
    if dest is None:
        dest = []
    if arr:
        dest.append(arr[0])
    for i in range(1, len(arr)):
        dest.append(delimiter)
        dest.append(arr[i])
    if at_end:
        dest.append(delimiter)
    return dest

def delimited_multi_line(node, text, begin=None, end=None, add_space=False):
    """Used to deal with single and multi line literals."""
    begin = begin or ''
    end = end or ''
    if begin and not end:
        end = begin
    sp = ' ' if add_space else ''
    lines = text.splitlines()
    if len(lines) > 1:
        yield node.line(node.part(begin, lines[0].strip()))
        for l in lines[1:-1]:
            yield node.line(l.strip())
        yield node.line(node.part(lines[-1].strip(), end))
    else:
        yield node.part(begin, sp, text, sp, end)


def parent_of(path):
    return '/'.join(path.rstrip('/').split('/')[:-1])


def body_top_names(body):
    names = set()
    for x in body:
        names |= node_names(x)
    return names


def controlled_ast_walk(node):
    """Walk AST just like ``ast.walk()``, but expect ``True`` on every branch to
    descend on sub-branches."""
    if isinstance(node, list):
        l = node.copy()
    elif isinstance(node, tuple):
        l = list(node)
    else:
        l = [node]
    while len(l) > 0:
        popped = l.pop()
        check_children = (yield popped)
        if check_children:
            for n in ast.iter_child_nodes(popped):
                l.append(n)

CODE_BLOCK_STMTS = (ast.FunctionDef, ast.ClassDef,
                    ast.AsyncFunctionDef)


def walk_under_code_boundary(node):
    it = controlled_ast_walk(node)
    traverse = None
    try:
        while True:
            subn = it.send(traverse)
            yield subn
            if isinstance(subn, CODE_BLOCK_STMTS):
                traverse = False # continue traversing sub names
            else:
                traverse = True
    except StopIteration:
        pass


def body_local_names(body):
    """Find the names assigned to in the provided `body`. It doesn't descend
    into function or class subelements."""
    names = set()
    for stmt in body:
        for subn in walk_under_code_boundary(stmt):
            if not isinstance(subn, CODE_BLOCK_STMTS):
                names |= node_names(subn)
    return names


def node_names(py_node):
    """Extract 'names' from a Python node. Names are all those interesting for the
    enclosing scope.

    Return a set containing them. The nodes considered are the Assign and the
    ones that defines namespaces, the function and class definitions.

    The assignment can be something like:

    .. code:: python
      a = b # 'a' is the target
      a = b = c # 'a' and 'b' are the targets
      a1, a2 = b = c # ('a1', 'a2') and 'b' are the targets
    """
    names = set()
    if isinstance(py_node, ast.Assign):
        for el in py_node.targets:
            if isinstance(el, ast.Name) and el.id not in \
               IGNORED_NAMES:
                names.add(el.id)
            elif isinstance(el, ast.Tuple):
                for elt in el.elts:
                    if isinstance(elt, ast.Name) and elt.id not in \
                       IGNORED_NAMES:
                        names.add(elt.id)
    elif isinstance(py_node, (ast.FunctionDef, ast.ClassDef)):
        names.add(py_node.name)
    return names


def rfilter(r, it, invert=False):
    """
    >>> list(rfilter(r'^.o+$', ['foo', 'bar']))
    ['foo']

    >>> list(rfilter(r'^.o+$', ['foo', 'bar'], invert=True))
    ['bar']

    """
    # Supports Python 2 and 3
    if isinstance(r, str):
        r = re.compile(r)
    try:
        if isinstance(r, unicode):
            r = re.compile
    except NameError:
        pass

    for x in it:
        m = r.search(x)
        ok = False
        if m:
            ok = True
        if invert:
            if not ok:
                yield x
        else:
            if ok:
                yield x


class OutputSrc:

    def __init__(self, node, name=None):
        self.node = node
        self.src_name = name

    def _gen_mapping(self, text, src_line=None, src_offset=None, dst_offset=None):
        """Generate a single mapping. `dst_line` is absent from signature
        because the part hasn't this information, but is present in the
        returned mapping. `src_line` is adjusted to be 0-based.

        See `Source Map version 3 proposal
        <https://docs.google.com/document/d/1U1RGAehQwRypUTovF1KRlpiOFze0b-_2gc6fAH0KY0k>`_.
        """
        return {
            'src_line': src_line,
            'src_offset': src_offset,
            'dst_line': None,
            'dst_offset': dst_offset,
            'text': text,
            'name': self.src_name if self.src_name is not True else str(self),
            'part': self
        }

    def _pos_in_src(self):
        """Return the position in source of the generated node."""
        py_node = self.node.py_node
        if py_node:
            offset = getattr(py_node, 'col_offset', 0)
            # multi-line comments have an offset of -1
            if offset < 0:
                offset = 0

            # special handling of nodes that are decorable. Those nodes expose
            # a 'lineno' that starts with the first decorator. for now, take
            # the last decorator lineno and add one
            if isinstance(py_node, (ast.FunctionDef, ast.AsyncFunctionDef,
                                    ast.ClassDef)) and py_node.decorator_list:
                result = (py_node.decorator_list[-1].lineno + 1, offset)
            else:
                result = (getattr(py_node, 'lineno', None),
                          offset)
        else:
            result = (None, None)
        return result


class Line(OutputSrc):

    def __init__(self, node, item, indent=False, delim=False, name=None):
        super().__init__(node, name)
        self.indent = int(indent)
        self.delim = delim
        if isinstance(item, (tuple, list)):
            item = Part(node, *item)
        self.item = item

    def __str__(self):
        line = str(self.item)
        if self.delim:
            line += ';'
        if self.indent and line.strip():
            line = (' ' * 4 * self.indent) + line
        line += '\n'
        return line

    def serialize(self):
        yield self

    def src_mappings(self):
        if self.node.transformer.disable_srcmap:
            return
        src_line, src_offset = self._pos_in_src()
        offset = self.indent * 4
        if isinstance(self.item, str):
            if src_line:
                yield self._gen_mapping(self.item, src_line, src_offset,
                                        offset)
        else:
            assert isinstance(self.item, Part)
            for m in self.item.src_mappings():
                m['dst_offset'] += offset
                yield m

    def __repr__(self):
        return '<%s indent: %d, "%s">' % (self.__class__.__name__,
                                          self.indent, str(self))


class Part(OutputSrc):

    def __init__(self, node, *items, name=None):
        super().__init__(node, name)
        self.items = []
        for i in items:
            if isinstance(i, (str, Part)):
                self.items.append(i)
            elif inspect.isgenerator(i):
                self.items.extend(i)
            else:
                self.items.append(str(i))

    def __str__(self):
        return ''.join(str(i) for i in self.items)

    def serialize(self):
        yield self

    def src_mappings(self):
        if self.node.transformer.disable_srcmap:
            return
        # optional position in source file, if this is missing, there's no
        # reason for generate a source mapping. (not all python AST elements
        # can be source located)
        src_line, src_offset = self._pos_in_src()
        # accumulator for string text
        frag = ''
        col = 0
        # for every item that composes this part...
        for i in self.items:
            assert isinstance(i, (str, Part))
            if isinstance(i, str):
                # if it's a string, just add it to the accumulator (usually
                # comma, parens, etc...)
                frag += i
            elif isinstance(i, Part):
                # if the item is a part
                if frag and src_line:
                    # .. and if there is accumulated text and a src location
                    # emit a src mapping for the accumulated text and reset it
                    yield self._gen_mapping(frag, src_line, src_offset, col)
                col += len(frag)
                frag = ''
                psrc = str(i)
                # ... then , ask the subpart to produce a src mapping and
                # maybe relocate it if necessary
                # yield from i._translate_src_mappings(i, src, psrc, col)
                for m in i.src_mappings():
                    m['dst_offset'] += col
                    yield m
                col += len(psrc)
        else:
            # at the end of the loop, if there is still a fragment and a src
            # location, emit a mapping for it
            if frag and src_line:
                yield self._gen_mapping(frag, src_line, src_offset, col)

    def _translate_src_mappings(self, part, src=None, psrc=None, start=None):
        src = src or str(self)
        psrc = psrc or str(part)
        offset = src.find(psrc, start)
        for m in part.src_mappings():
            m['dst_offset'] += offset
            yield m

    def __repr__(self):
        return '<%s, "%s">' % (self.__class__.__name__,
                               str(self))


def linecounter(iterable, start=1):
    count = start
    for line in iterable:
        yield count, line
        count += len(re.findall('\n', str(line)))


class Block(OutputSrc):

    def __init__(self, node):
        super().__init__(None)
        self.lines = list(node.serialize())

    def src_mappings(self, src_offset=None, dst_offset=None):

        sline_offset, scol_offset = src_offset or (0, 0)
        dline_offset, dcol_offset = dst_offset or (0, 0)
        for ix, line in linecounter(self.lines, start=1):
            for m in line.src_mappings():
                m['dst_line'] = ix + dline_offset
                m['dst_offset'] += dcol_offset
                m['src_line'] += sline_offset
                m['src_offset'] += scol_offset
                yield m

    def read(self):
        return ''.join(str(l) for l in self.lines)

    def sourcemap(self, source, src_filename, src_offset=None, dst_offset=None):
        Token = sourcemaps.Token
        tokens = []
        for m in self.src_mappings(src_offset, dst_offset):
            token = Token(m['dst_line'] - 1, m['dst_offset'], src_filename,
                          m['src_line'] - 1, m['src_offset'], m['name'], m)
            tokens.append(token)

        src_map = sourcemaps.SourceMap(
            sources_content={src_filename: source}
        )
        for t in tokens:
            src_map.add_token(t)
        return sourcemaps.encode(src_map), src_map


def obj_source(obj):
    src = inspect.getsource(obj)
    src = textwrap.dedent(src)
    return src



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- common transformation functions
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
#            Devan Lai <devan.lai@gmail.com>
# :License:  GNU General Public License version 3 or later
#

import ast
from ..js_ast import (
    JSBinOp,
    JSMultipleArgsOp,
    JSName,
    JSOpInstanceof,
    JSOpOr,
    JSOpStrongEq,
    JSOpTypeof,
    JSStr,
    JSUnaryOp,
)


def _build_call_isinstance(tgt, cls_or_seq):
    """Helper to build the translate the equivalence of ``isinstance(foo, Bar)``
    to ``foo instanceof Bar`` and ``isinstance(Foo, (Bar, Zoo))`` to
    ``foo instanceof Bar || foo instanceof Zoo``.
    """
    if isinstance(cls_or_seq, (ast.Tuple, ast.List, ast.Set)):
        classes = cls_or_seq.elts
        args = tuple((tgt, c) for c in classes)
        return JSMultipleArgsOp(JSOpInstanceof(), JSOpOr(), *args)
    else:
        cls = cls_or_seq
        if isinstance(cls, ast.Name) and cls.id == 'str':
            return JSMultipleArgsOp(
                (JSOpStrongEq(), JSOpInstanceof()),
                JSOpOr(),
                (JSUnaryOp(JSOpTypeof(), tgt), JSStr('string')),
                (tgt, JSName('String'))
            )
        elif isinstance(cls, ast.Name) and cls.id in ['int', 'float']:
            return JSMultipleArgsOp(
                (JSOpStrongEq(), JSOpInstanceof()),
                JSOpOr(),
                (JSUnaryOp(JSOpTypeof(), tgt), JSStr('number')),
                (tgt, JSName('Number'))
            )
        else:
            return JSBinOp(tgt, JSOpInstanceof(), cls)



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- function transformations
# :Created:  mer 09 nov 2016 12:59:00 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
from functools import reduce
from itertools import chain

from ..js_ast import (
    JSAttribute,
    JSArrowFunction,
    JSAssignmentExpression,
    JSAsyncFunction,
    JSAsyncMethod,
    JSCall,
    JSClassConstructor,
    JSDict,
    JSExpressionStatement,
    JSFunction,
    JSGenFunction,
    JSGenMethod,
    JSGetter,
    JSMethod,
    JSName,
    JSRest,
    JSSetter,
    JSStatements,
    JSThis,
    JSVarStatement,
)
from ..processor.util import body_local_names, walk_under_code_boundary

from . import _normalize_name


def _isyield(el):
    return isinstance(el, (ast.Yield, ast.YieldFrom))


def FunctionDef(t, x, fwrapper=None, mwrapper=None):

    is_method = isinstance(t.parent_of(x), ast.ClassDef)
    is_in_method = not x.name.startswith('fn_') and \
                   all(lambda p: isinstance(p, (ast.FunctionDef,
                                                ast.AsyncFunctionDef,
                                                ast.ClassDef)) \
                       for p in t.parents(x, stop_at=ast.ClassDef)) and \
                           isinstance(tuple(t.parents(x, stop_at=ast.ClassDef))[-1],
                                      ast.ClassDef) # Make sure a class is there

    is_generator = reduce(lambda prev, cur: _isyield(cur) or prev,
                          walk_under_code_boundary(x.body), False)

    t.unsupported(x, not is_method and x.decorator_list, "Function decorators are"
                  " unsupported yet")

    t.unsupported(x, len(x.decorator_list) > 1, "No more than one decorator"
                  " is supported")

    if x.args.vararg or x.args.kwonlyargs or x.args.defaults or \
       x.args.kw_defaults or x.args.kwarg:
        t.es6_guard(x, "Arguments definitions other than plain params require "
                    "ES6 to be enabled")

    t.unsupported(x, x.args.kwarg and x.args.kwonlyargs,
                  "Keyword arguments together with keyword args accumulator"
                  " are unsupported")

    t.unsupported(x, x.args.vararg and (x.args.kwonlyargs or x.args.kwarg),
                  "Having both param accumulator and keyword args is "
                  "unsupported")

    name = _normalize_name(x.name)
    body = x.body
    # get positional arg names and trim self if present
    arg_names = [arg.arg for arg in x.args.args]
    if is_method or (len(arg_names) > 0 and arg_names[0] == 'self'):
        arg_names = arg_names[1:]


    acc = JSRest(x.args.vararg.arg) if x.args.vararg else None
    defaults = x.args.defaults
    kw = x.args.kwonlyargs
    kwdefs = x.args.kw_defaults
    kw_acc = x.args.kwarg
    kw_names = [k.arg for k in kw]
    if kw:
        kwargs = []
        for k, v in zip(kw, kwdefs):
            if v is None:
                kwargs.append(k.arg)
            else:
                kwargs.append(JSAssignmentExpression(k.arg, v))
    else:
        kwargs = None

    # be sure that the defaults equal in length the args list
    if isinstance(defaults, (list, tuple)) and len(defaults) < len(arg_names):
        defaults = ([None] * (len(arg_names) - len(defaults))) + list(defaults)
    elif defaults is None:
        defaults = [None] * len(arg_names)

    if kw_acc:
        arg_names += [kw_acc.arg]
        defaults += [JSDict((), ())]

    # render defaults of positional arguments and keywords accumulator
    args = []
    for k, v in zip(arg_names, defaults):
        if v is None:
            args.append(k)
        else:
            args.append(JSAssignmentExpression(k, v))

    # local function vars
    if 'vars' in t.ctx:
        upper_vars = t.ctx['vars']
    else:
        upper_vars = set()
    local_vars = list((set(body_local_names(body)) - set(arg_names)) -
                      set(kw_names) - upper_vars)
    t.ctx['vars'] = upper_vars | set(local_vars)
    if len(local_vars) > 0:
        local_vars.sort()
        body = JSStatements(
            JSVarStatement(local_vars, [None] * len(local_vars)),
            *body
        )

    if is_generator:
        fwrapper = JSGenFunction
        mwrapper = JSGenMethod

    # If x is a method
    if is_method:
        cls_member_opts = {}
        if x.decorator_list:
            # decorator should be "property" or "<name>.setter" or "classmethod"
            fdeco = x.decorator_list[0]
            if isinstance(fdeco, ast.Name) and fdeco.id == 'property':
                deco = JSGetter
            elif (isinstance(fdeco, ast.Attribute) and  fdeco.attr == 'setter'
                  and isinstance(fdeco.value, ast.Name)):
                deco = JSSetter
            elif isinstance(fdeco, ast.Name) and fdeco.id == 'classmethod':
                deco = None
                cls_member_opts['static'] = True
            else:
                t.unsupported(x, True, "Unsupported method decorator")
        else:
            deco = None

        if name == '__init__':
            result = JSClassConstructor(
                args, body, acc, kwargs
            )
        else:
            mwrapper = mwrapper or deco or JSMethod
            if mwrapper is JSGetter:
                result = mwrapper(
                    name, body,
                    **cls_member_opts
                )
            elif mwrapper is JSSetter:
                t.unsupported(x, len(args) == 0, "Missing argument in setter")
                result = mwrapper(
                    name, args[0], body,
                    **cls_member_opts
                )
            elif mwrapper is JSMethod:
                if name == '__len__':
                    result = JSGetter(
                        'length',
                        body,
                        **cls_member_opts
                    )
                elif name == '__str__':
                    result = JSMethod(
                        'toString',
                        [], body,
                        **cls_member_opts
                    )
                elif name == '__get__':
                    result = JSMethod(
                        'get',
                        [], body,
                        **cls_member_opts
                    )
                elif name == '__set__':
                    result = JSMethod(
                        'set',
                        [], body,
                        **cls_member_opts
                    )
                elif name == '__instancecheck__':
                    cls_member_opts['static'] = True
                    result = JSMethod(
                        '[Symbol.hasInstance]',
                        args, body,
                        **cls_member_opts
                    )
                else:
                    result = mwrapper(
                        name, args, body,
                        acc, kwargs,
                        **cls_member_opts
                    )
            else:
                result = mwrapper(
                    name, args, body,
                    acc, kwargs,
                    **cls_member_opts
                )
    # x is a function
    else:
        if is_in_method and fwrapper is None:
            # set the incoming py_node for the sourcemap
            fdef = JSArrowFunction(
                name, args, body, acc, kwargs
            )
            fdef.py_node = x
            result = JSStatements(
                JSVarStatement([str(name)], [None]),
                fdef
            )
        elif is_in_method and fwrapper in [JSGenFunction, JSAsyncFunction]:
            # set the incoming py_node for the sourcemap
            fdef = fwrapper(
                name, args, body, acc, kwargs
            )
            fdef.py_node = x
            # arrow functions cannot be generators, render them as normal
            # function and add a bind(self)
            result = JSStatements(
                fdef,
                JSExpressionStatement(
                    JSAssignmentExpression(
                        JSName(name),
                        JSCall(
                            JSAttribute(name, 'bind'),
                            [JSThis()]
                        )
                    )
                )
            )
        else:
            fwrapper = fwrapper or JSFunction
            result = fwrapper(
                name, args, body,
                acc, kwargs
            )
    return result


def AsyncFunctionDef(t, x):
    t.stage3_guard(x, "Async stuff requires 'stage3' to be enabled")
    return FunctionDef(t, x, JSAsyncFunction, JSAsyncMethod)



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- special transformations
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
import re

from ..js_ast import (
    JSAssignmentExpression,
    JSAttribute,
    JSBinOp,
    JSCall,
    JSCommentBlock,
    JSDependImport,
    JSDict,
    JSExpressionStatement,
    JSLiteral,
    JSName,
    JSNamedImport,
    JSNull,
    JSNum,
    JSOpIn,
    JSOpInstanceof,
    JSOpNot,
    JSOpOr,
    JSOpStrongEq,
    JSOpStrongNotEq,
    JSOpTypeof,
    JSPass,
    JSStarImport,
    JSStatements,
    JSStr,
    JSSubscript,
    JSTaggedTemplate,
    JSTemplateLiteral,
    JSThis,
    JSUnaryOp,
)

from .classes import (
    Attribute_super,
    Call_isinstance,
    Call_issubclass,
    Call_super,
    Subscript_super,
)

from .obvious import (
    Attribute_default,
    BinOp_default,
    Call_default,
    Compare_default,
    Expr_default,
    Name_default,
    Subscript_default,
)

from . import _normalize_name


#### Expr

# docstrings &rarr; comment blocks
def Expr_docstring(t, x):
    if isinstance(x.value, ast.Str):
        return JSCommentBlock(x.value.s)

Expr = [Expr_docstring, Expr_default]


# <code>2**3</code> &rarr; <code>Math.pow(2, 3)</code>
def BinOp_pow(t, x):
    if isinstance(x.op, ast.Pow):
        return JSCall(
            JSAttribute(
                JSName('Math'),
                'pow'),
            [x.left, x.right])


BinOp = [BinOp_pow, BinOp_default]


# <code>self</code> &rarr; <code>this</code>
def Name_self(t, x):
    if x.id == 'self':
        return JSThis()


Name = [Name_self, Name_default]


# <code>typeof(x)</code> &rarr; <code>(typeof x)</code>
def Call_typeof(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == 'typeof'):
        assert len(x.args) == 1
        return JSUnaryOp(JSOpTypeof(), x.args[0])


def Call_callable(t, x):
    """Translate ``callable(foo)`` to ``foo instanceof Function``."""
    if (isinstance(x.func, ast.Name) and x.func.id == 'callable'):
        assert len(x.args) == 1
        return JSBinOp(
            JSBinOp(x.args[0], JSOpInstanceof(), JSName('Function')),
            JSOpOr(),
            JSBinOp(
                JSUnaryOp(JSOpTypeof(), x.args[0]),
                JSOpStrongEq(),
                JSStr('function')
            )
        )


# <code>print(...)</code> &rarr; <code>console.log(...)</code>
def Call_print(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == 'print'):
        return JSCall(JSAttribute(JSName('console'), 'log'), x.args)


# <code>len(x)</code> &rarr; <code>x.length</code>
def Call_len(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == 'len' and
        len(x.args) == 1):
        return JSAttribute(x.args[0], 'length')


def Call_str(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == 'str' and
        len(x.args) == 1):
        return JSCall(JSAttribute(JSName(x.args[0]), 'toString'), [])


def Call_new(t, x):
    """Translate ``Foo(...)`` to ``new Foo(...)`` if function name starts
    with a capital letter.
    """
    def getNameString(x):
        if isinstance(x, ast.Name):
            return x.id
        elif isinstance(x, ast.Attribute):
            return str(x.attr)
        elif isinstance(x, ast.Subscript):
            if isinstance(x.slice, ast.Index):
                return str(x.slice.value)

    NAME_STRING = getNameString(x.func)

    if (NAME_STRING and re.search(r'^[A-Z]', NAME_STRING)):
        # TODO: generalize args mangling and apply here
        # assert not any([x.keywords, x.starargs, x.kwargs])
        subj = x
    elif isinstance(x.func, ast.Name) and x.func.id == 'new':
        subj = x.args[0]
    else:
        subj = None
    if subj:
        return Call_default(t, subj, operator='new ')


def Call_import(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == '__import__'):
        assert len(x.args) == 1 and isinstance(x.args[0], ast.Str)
        t.es6_guard(x, "'__import__()' call requires ES6")
        return JSDependImport(x.args[0].s)


def Call_type(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == 'type'):
        assert len(x.args) == 1
        return JSCall(JSAttribute(JSName('Object'), 'getPrototypeOf'), x.args)


def Call_dict_update(t, x):
    """Convert ``dict(foo).update(bar)`` to ``Object.assign(foo, bar)``.

    Requires ES6

    AST dump::

      Expr(value=Call(args=[Name(ctx=Load(),
                                 id='bar')],
                      func=Attribute(attr='update',
                                     ctx=Load(),
                                     value=Call(args=[Name(ctx=Load(),
                                                           id='foo')],
                                                func=Name(ctx=Load(),
                                                          id='dict'),
                                                keywords=[])),
                      keywords=[]))

    """
    if isinstance(x.func, ast.Attribute) and x.func.attr == 'update' and \
       isinstance(x.func.value, ast.Call) and  \
       isinstance(x.func.value.func, ast.Name) and \
       x.func.value.func.id == 'dict' and  len(x.func.value.args) == 1:
        t.es6_guard(x, "dict.update() requires ES6")
        return JSCall(
            JSAttribute(JSName('Object'), 'assign'),
            [x.func.value.args[0]] + x.args
            )


def Call_dict_copy(t, x):
    """Convert ``dict(foo).copy()`` to ``Object.assign({}, foo)``.

    Requires ES6

    AST dump::

      Expr(value=Call(args=[],
                      func=Attribute(attr='copy',
                                     ctx=Load(),
                                     value=Call(args=[Name(ctx=Load(),
                                                           id='foo')],
                                                func=Name(ctx=Load(),
                                                          id='dict'),
                                                keywords=[])),
                      keywords=[]))
    """
    if isinstance(x.func, ast.Attribute) and x.func.attr == 'copy' and \
       isinstance(x.func.value, ast.Call) and  \
       isinstance(x.func.value.func, ast.Name) and \
       x.func.value.func.id == 'dict' and len(x.func.value.args) == 1:
        t.es6_guard(x, "dict.copy() requires ES6")
        return JSCall(
            JSAttribute(JSName('Object'), 'assign'),
            (JSDict([], []), x.func.value.args[0])
            )


def Call_template(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == 'tmpl') and \
       len(x.args) > 0:
        assert len(x.args) == 1
        assert isinstance(x.args[0], ast.Str)
        t.es6_guard(x, "template literals require ES6")
        return JSTemplateLiteral(x.args[0].s)


def Call_tagged_template(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == '__') and \
       len(x.args) > 0 and t.parent_of(x) is not ast.Attribute:
        assert 3 > len(x.args) >= 1
        assert isinstance(x.args[0], ast.Str)
        t.es6_guard(x, "tagged templates require ES6")
        if len(x.args) == 2:
            tag = x.args[1]
        else:
            tag = JSName('__')
        return JSTaggedTemplate(x.args[0].s, tag)


def Call_hasattr(t, x):
    """Translate ``hasattr(foo, bar)`` to ``bar in foo``."""
    if (isinstance(x.func, ast.Name) and x.func.id == 'hasattr') and \
       len(x.args) == 2:
        return JSBinOp(x.args[1], JSOpIn(), x.args[0])


def Call_getattr(t, x):
    """Translate ``getattr(foo, bar, default)`` to ``foo[bar] || default``."""
    if (isinstance(x.func, ast.Name) and x.func.id == 'getattr') and \
       2 <= len(x.args) < 4:
        if len(x.args) == 2:
            res = JSSubscript(x.args[0], x.args[1])
        else:
            res = JSBinOp(
                JSSubscript(x.args[0], x.args[1]),
                JSOpOr(),
                x.args[2]
            )
        return res


def Call_setattr(t, x):
    """Translate ``setattr(foo, bar, value)`` to ``foo[bar] = value``."""
    if (isinstance(x.func, ast.Name) and x.func.id == 'setattr') and \
       len(x.args) == 3:
        return JSExpressionStatement(
            JSAssignmentExpression(
                JSSubscript(x.args[0], x.args[1]),
                x.args[2]
            )
        )


def Call_JS(t, x):
    if (isinstance(x.func, ast.Name) and x.func.id == 'JS') and \
       len(x.args) == 1:
        assert isinstance(x.args[0], ast.Str)
        return JSLiteral(x.args[0].s)


def Call_int(t, x):
    # maybe this needs a special keywords mangling for optional "base" param
    if isinstance(x.func, ast.Name) and x.func.id == 'int':
        if t.enable_es6:
            return JSCall(JSAttribute('Number', 'parseInt'), x.args)
        else:
            return JSCall(JSName('parseInt'), x.args)


def Call_float(t, x):
    if isinstance(x.func, ast.Name) and x.func.id == 'float':
        if t.enable_es6:
            return JSCall(JSAttribute('Number', 'parseFloat'), x.args)
        else:
            return JSCall(JSName('parseFloat'), x.args)


Call = [Call_typeof, Call_callable, Call_isinstance, Call_print, Call_len,
        Call_JS, Call_new, Call_super, Call_import, Call_str, Call_type,
        Call_dict_update, Call_dict_copy, Call_tagged_template, Call_template,
        Call_hasattr, Call_getattr, Call_setattr, Call_issubclass,
        Call_int, Call_float, Call_default]


#### Ops

# <code>==</code>
#
# Transform to <code>===</code>
def Eq(t, x):
    return JSOpStrongEq()

Is = Eq

# <code>!=</code>
#
# Transform to <code>!==</code>
def NotEq(t, x):
    return JSOpStrongNotEq()

IsNot = NotEq

#### Import


AT_PREFIX_RE = re.compile(r'^__([a-zA-Z0-9])')
INSIDE_DUNDER_RE = re.compile(r'([a-zA-Z0-9])__([a-zA-Z0-9])')

def _replace_dunder(name):
    """Replace dunder (``__``) in `name` with an ``@`` symbol if it's at the
    start and with ``-`` if it's on the middle."""
    res = AT_PREFIX_RE.sub(r'@\1', name)
    return INSIDE_DUNDER_RE.sub(r'\1-\2', res)


def Import(t, x):
    t.es6_guard(x, "'import' statement requires ES6")
    names = []
    for n in x.names:
        names.append(n.asname or n.name)
    t.add_globals(*names)
    result = []
    for n in x.names:
        old_name = n.name
        n.name = _replace_dunder(n.name)
        t.unsupported(x, (old_name != n.name) and not n.asname,
                      "A module name cannot contain dashes, use 'as' to give it a new name.")
        path_module = '/'.join(n.name.split('.'))
        result.append(
            JSStarImport(path_module, n.asname or n.name)
        )
    return JSStatements(*result)


def ImportFrom(t, x):
    names = []
    for n in x.names:
        names.append(n.asname or n.name)
    if x.module == '__globals__':
        assert x.level == 0
        # assume a fake import to import js stuff from root object
        t.add_globals(*names)
        result = JSPass()
    else:
        t.es6_guard(x, "'import from' statement requires ES6")
        t.add_globals(*names)
        result = JSPass()
        if x.module:
            mod = tuple(_normalize_name(frag) for frag in
                        _replace_dunder(x.module).split('.'))
            path_module = '/'.join(mod)
            if x.level == 1:
                # from .foo import bar
                path_module = './' + path_module
            elif x.level > 1:
                # from ..foo import bar
                # from ...foo import bar
                path_module = '../' * (x.level - 1) + path_module
            result = JSNamedImport(path_module,
                                   [(n.name, n.asname) for n in x.names])
        else:
            assert x.level > 0
            result = []
            for n in x.names:
                if x.level == 1:
                    # from . import foo
                    imp = JSStarImport('./' + n.name, n.asname or n.name)
                else:
                    # from .. import foo
                    imp = JSStarImport('../' * (x.level -1) + n.name,
                                       n.asname or n.name)
                if len(x.names) == 1:
                    imp.py_node = x
                else:
                    imp.py_node = n
                result.append(imp)
            result = JSStatements(*result)
    return result


def Compare_in(t, x):
    if not isinstance(x.ops[0], (ast.NotIn, ast.In)):
        return
    if t.enable_snippets:
        from ..snippets import _in, in_es6
        if t.enable_es6:
            t.add_snippet(in_es6)
            sname = 'in_es6'
        else:
            t.add_snippet(_in)
            sname = '_in'
        result = JSCall(JSAttribute('_pj', sname), [x.left, x.comparators[0]])
        if isinstance(x.ops[0], ast.NotIn):
            result = JSUnaryOp(JSOpNot(), result)
        return result


Compare = [Compare_in, Compare_default]


def Subscript_slice(t, x):

    if isinstance(x.slice, ast.Slice):
        slice = x.slice
        t.unsupported(x, slice.step and slice.step != 1, "Slice step is unsupported")
        args = []
        if slice.lower:
            args.append(slice.lower)
        else:
            args.append(JSNum(0))
        if slice.upper:
            args.append(slice.upper)

        return JSCall(JSAttribute(x.value, 'slice'), args)


Subscript = [Subscript_slice, Subscript_super, Subscript_default]


def Attribute_list_append(t, x):
    """Convert ``list(foo).append(bar)`` to ``foo.push(bar)``.

    AST dump::

      Expr(value=Call(args=[Name(ctx=Load(),
                                 id='bar')],
                      func=Attribute(attr='append',
                                     ctx=Load(),
                                     value=Call(args=[Name(ctx=Load(),
                                                           id='foo')],
                                                func=Name(ctx=Load(),
                                                          id='list'),
                                                keywords=[])),
                      keywords=[]))
    """
    if x.attr == 'append' and isinstance(x.value, ast.Call) and \
       isinstance(x.value.func, ast.Name) and x.value.func.id == 'list' and \
       len(x.value.args) == 1:
        return JSAttribute(x.value.args[0], 'push')


Attribute = [Attribute_super, Attribute_list_append, Attribute_default]


def Assert(t, x):
    """Convert ``assert`` statement to just a snippet function call."""
    if t.enable_snippets:
        from ..snippets import _assert
        t.add_snippet(_assert)
        return JSCall(JSAttribute('_pj', '_assert'), [x.test, x.msg or JSNull()])



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- JS AST
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>,
#            Lele Gaifax <lele@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import itertools
import json
import re

from .processor.transforming import TargetNode
from .processor.util import delimited, delimited_multi_line


JS_KEYWORDS = set([
    'break', 'case', 'catch', 'continue', 'default', 'delete', 'do', 'else',
    'finally', 'for', 'function', 'if', 'in', 'instanceof', 'new', 'return',
    'switch', 'this', 'throw', 'try', 'typeof', 'var', 'void', 'while', 'with',

    'abstract', 'boolean', 'byte', 'char', 'class', 'const',
    'double', 'enum', 'export', 'extends', 'final', 'float', 'goto',
    'implements', 'import', 'int', 'interface', 'long', 'native', 'package',
    'private', 'protected', 'public', 'short', 'static', 'super',
    'synchronized', 'throws', 'transient', 'volatile'])

JS_KEYWORDS_ES6 = JS_KEYWORDS - set(['delete'])


def _check_keywords(target_node, name):
    trans = target_node.transformer
    if trans is not None:
        trans.unsupported(
            target_node.py_node,
            name in JS_KEYWORDS_ES6 if trans.enable_es6 else name in JS_KEYWORDS,
            "Name '%s' is reserved in JavaScript." % name)
    else:
        if name in JS_KEYWORDS:
            raise ValueError("Name %s is reserved in JavaScript." % name)

#### Misc


class JSNode(TargetNode):
    pass


class JSStatement(JSNode):
    pass


class JSLeftSideUnaryOp(JSNode):
    pass


class JSStatements(JSNode):

    def __iadd__(self, other):
        self.transformed_args.extend(other.transformed_args)
        return self

    def emit(self, statements):
        for s in statements:
            yield s

    def squash(self, args):
        for a in args:
            if isinstance(a, JSStatements):
                yield from a.transformed_args
            else:
                yield a

    def reordered_args(self, args):
        """Reorder the args to keep the imports and vars always at the top."""
        args = list(self.squash(args))
        imports = []
        vars_ = []
        others = []
        for a in args:
            if isinstance(a, JSImport):
                imports.append(a)
            elif isinstance(a, JSVarStatement) and \
                 not a.options.get('unmovable', False):
                vars_.append(a)
            else:
                others.append(a)

        others_first = []
        others_after = []
        # if the others start with some comments, put those at the top
        start_trigger = False
        for s in others:
            if isinstance(s, JSCommentBlock) and not start_trigger:
                others_first.append(s)
            else:
                others_after.append(s)
                start_trigger = True

        return itertools.chain(others_first, imports, vars_, others_after)

    def serialize(self):
        for a in self.emit(self.reordered_args(self.transformed_args)):
            yield from self.lines(a.serialize(), delim=True)


class JSPass(JSNode):
    def emit(self):
        return []


class JSCommentBlock(JSNode):
    def emit(self, text):
        assert text.find('*/') == -1
        yield from self.lines(delimited_multi_line(self, text, '/*', '*/', True))


class JSLiteral(JSNode):
    def emit(self, text):
        yield from self.lines(delimited_multi_line(self, text, '', '', False))

#### Statements


class JSExpressionStatement(JSStatement):
    def emit(self, value):
        yield self.part(value)


class JSVarDeclarer(JSStatement):

    def with_kind(self, kind, keys, values):
        for key in keys:
            _check_keywords(self, key)
        assert len(keys) > 0
        assert len(keys) == len(values)

        arr = ['%s ' % kind]
        for i in range(len(keys)):
            if i > 0:
                arr.append(', ')
            arr.append(keys[i])
            if values[i] is not None:
                arr.append(' = ')
                arr.append(values[i])
        yield self.part(*arr)


class JSVarStatement(JSVarDeclarer):
    def emit(self, keys, values, unmovable=False):
        yield from self.with_kind('var', keys, values)


class JSLetStatement(JSVarDeclarer):
    def emit(self, keys, values, unmovable=True):
        yield from self.with_kind('let', keys, values)


class JSAugAssignStatement(JSStatement):
    def emit(self, target, op, value):
        yield self.part(target, ' ', op, '= ', value, name=str(target))


class JSIfStatement(JSStatement):
    def emit(self, test, body, orelse):
        yield self.line(['if (', test, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        if orelse:
            yield self.line(['} else {'])
            yield from self.lines(orelse, indent=True, delim=True)
            yield self.line('}')
        else:
            yield self.line('}')


class JSWhileStatement(JSStatement):
    def emit(self, test, body):
        yield self.line(['while (', test, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSForStatement(JSStatement):
    def emit(self, left, test, right, body):
        yield self.line(['for (', left, '; ', test, '; ', right, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSForIterableStatement(JSStatement):

    operator = ' of '

    def emit(self, target, source, body):
        yield self.line(['for (var ', self.part(target), self.operator,
                         source, ') {'])
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSForeachStatement(JSForIterableStatement):

    operator = ' in '


class JSForofStatement(JSForIterableStatement):
    pass


class JSReturnStatement(JSStatement):
    def emit(self, value):
        if value:
            result = self.line(['return ', value], delim=True)
        else:
            result = self.line('return', delim=True)
        yield result


class JSBreakStatement(JSStatement):
    def emit(self):
        yield self.part('break')


class JSContinueStatement(JSStatement):
    def emit(self):
        yield self.part('continue')


class JSDeleteStatement(JSStatement):
    def emit(self, value):
        yield self.line(['delete ', value], delim=True)


class JSTryCatchFinallyStatement(JSStatement):
    def emit(self, try_body, target, catch_body, finally_body):
        assert catch_body or finally_body
        yield self.line('try {')
        yield from self.lines(try_body, indent=True, delim=True)
        if catch_body:
            yield self.line(['} catch(', target, ') {'])
            yield from self.lines(catch_body, indent=True, delim=True)
        if finally_body:
            yield self.line(['} finally {'])
            yield from self.lines(finally_body, indent=True, delim=True)
        yield self.line('}')


class JSThrowStatement(JSStatement):
    def emit(self, obj):
        yield self.line(['throw ', obj], delim=True)


class JSImport(JSStatement):
    pass


class JSDependImport(JSImport):
    def emit(self, module):
        yield self.line(['System.import(', "'", module, "'", ')'], delim=True)


class JSNamedImport(JSImport):
    def emit(self, module, names):
        js_names = []
        for name, alias in sorted(names):
            if alias:
                js_names.append(self.part(name, ' as ', alias))
            else:
                js_names.append(self.part(name))

        yield self.line(['import {', *delimited(', ', js_names),
                         "} from '", module, "'"], delim=True)


class JSStarImport(JSImport):
    def emit(self, module, name):
        yield self.line(['import * as ', name, " from '", module, "'"],
                        delim=True)


class JSExport(JSStatement):
    def emit(self, name):
        yield self.line(['export ', '{', name, '}'], delim=True)


class JSAwait(JSStatement):

    def emit(self, value):
        yield self.part('await ', value)


class JSFunction(JSStatement):

    begin = 'function '
    bet_args_n_body = ''

    def fargs(self, args, acc=None, kwargs=None):
        result = []
        result.append('(')
        js_args = args.copy()
        if kwargs:
            js_args.append(self.part('{', *delimited(', ', kwargs), '}={}'))
        if acc:
            js_args.append(acc)
        delimited(', ', js_args, dest=result)
        result.append(') ')
        return result

    def emit(self, name, args, body, acc=None, kwargs=None):
        line = [self.begin]
        if name is not None:
            line.append(name)
        line += self.fargs(args, acc, kwargs)
        line += self.bet_args_n_body
        line += ['{']
        yield self.line(line, name=str(name))
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSAsyncFunction(JSFunction):

    begin = 'async function '


class JSGenFunction(JSFunction):

    begin = 'function* '


class JSArrowFunction(JSFunction):

    begin = ''
    bet_args_n_body = '=> '

    def emit(self, name, args, body, acc=None, kwargs=None):
        if name:
            # TODO: split this into an assignment + arrow function
            line = [name, ' = ']
        else:
            line = []
        line += self.fargs(args, acc, kwargs)
        line += self.bet_args_n_body
        line += ['{']
        yield self.line(line)
        yield from self.lines(body, indent=True, delim=True)
        if name:
            yield self.line('}', delim=True)
        else:
            yield self.part('}')


class JSClass(JSStatement):

    def emit(self, name, super_, methods):
        line = ['class ', name]
        if super_ is not None:
            line += [' extends ', super_]
        line += [' {']
        yield self.line(line)
        yield from self.lines(methods, indent=True, delim=True)
        yield self.line('}')


class JSClassMember(JSFunction):

    def with_kind(self, kind, args, body, acc=None, kwargs=None, static=False):
        if static:
            line = ['static ', kind]
        else:
            line = [kind]
        line += self.fargs(args, acc, kwargs)
        line += ['{']
        yield self.line(line)
        yield from self.lines(body, indent=True, delim=True)
        yield self.line('}')


class JSClassConstructor(JSClassMember):

    def emit(self, args, body, acc=None, kwargs=None):
        yield from self.with_kind('constructor', args, body, acc, kwargs)


class JSMethod(JSClassMember):

    def emit(self, name, args, body, acc=None, kwargs=None, static=False):
        yield from self.with_kind(name, args, body, acc, kwargs, static)


class JSAsyncMethod(JSClassMember):

    def emit(self, name, args, body, acc=None, kwargs=None, static=False):
        yield from self.with_kind('async ' + name, args, body, acc, kwargs,
                                  static)


class JSGenMethod(JSClassMember):

    def emit(self, name, args, body, acc=None, kwargs=None, static=False):
        yield from self.with_kind('* ' + name, args, body, acc, kwargs,
                                  static)


class JSGetter(JSClassMember):

    def emit(self, name, body, static=False):
        yield from self.with_kind('get ' + name, [], body, static=static)


class JSSetter(JSClassMember):

    def emit(self, name, arg, body, static=False):
        yield from self.with_kind('set ' + name, [arg], body, static=static)


#### Expressions


class JSList(JSNode):
    def emit(self, elts):
        arr = ['[']
        delimited(', ', elts, dest=arr)
        arr.append(']')
        yield self.part(*arr)


class JSDict(JSNode):
    def emit(self, keys, values):
        arr = ['{']
        for i in range(len(keys)):
            if i > 0:
                arr.append(', ')
            arr.append(keys[i])
            arr.append(': ')
            arr.append(values[i])
        arr.append('}')
        yield self.part(*arr)


class JSAssignmentExpression(JSNode):
    def emit(self, left, right):
        yield self.part(left, ' = ', right)


class JSIfExp(JSNode):
    def emit(self, test, body, orelse):
        yield self.part('(', test, ' ? ', body, ' : ', orelse, ')')


class JSCall(JSNode):

    operator = ''

    def emit(self, func, args, kwargs=None, operator=None):
        operator = operator or self.operator
        kwargs = kwargs or []
        arr = [operator, func, '(']
        fargs = args.copy()
        if kwargs:
            fargs.append(kwargs)
        delimited(', ', fargs, dest=arr)
        arr.append(')')
        yield self.part(*arr)


class JSNewCall(JSCall):

    operator = 'new '


class JSAttribute(JSNode):
    def emit(self, obj, s):
        assert re.search(r'^[a-zA-Z$_][a-zA-Z$_0-9]*$', s)
        _check_keywords(self, s)
        yield self.part(obj, '.', s, name=True)


class JSSubscript(JSNode):
    def emit(self, obj, key):
        yield self.part(self.part(obj, name=True), '[',
                        self.part(key, name=True), ']')


class JSKeySubscript(JSNode):
    def emit(self, key):
        yield self.part('[', self.part(key), ']')


class JSBinOp(JSNode):
    def emit(self, left, op, right):
        yield self.part('(', left, ' ', op, ' ', right, ')')


class JSMultipleArgsOp(JSNode):
    def emit(self, binop, conj, *args):
        assert len(args) > 1
        parts = []
        for ix, arg in enumerate(args):
            if isinstance(binop, (tuple, list)):
                op = binop[ix]
            else:
                op = binop
            if ix > 0:
                parts += [' ', conj, ' ']
            parts += ['(', arg[0], ' ', op, ' ', arg[1], ')']
        yield self.part('(', *parts, ')')


class JSUnaryOp(JSNode):
    def emit(self, op, right):
        assert isinstance(op, JSLeftSideUnaryOp)
        yield self.part('(', op, ' ', right, ')')


class JSYield(JSNode):

    def emit(self, expr):
        yield self.part('yield ', expr)


class JSYieldStar(JSNode):

    def emit(self, expr):
        yield self.part('yield* ', expr)


class JSExpression(JSNode):

    def emit(self, expr):
        yield self.part('(', expr, ')')


#### Atoms


class JSNum(JSNode):
    def emit(self, x):
        yield self.part(str(x))


class JSStr(JSNode):
    def emit(self, s):
        yield self.part(json.dumps(s))


class JSName(JSNode):
    def emit(self, name):
        _check_keywords(self, name)
        yield self.part(name, name=True)


class JSSuper(JSNode):
    def emit(self):
        yield self.part('super')


class JSThis(JSNode):
    def emit(self):
        yield self.part('this')


class JSTrue(JSNode):
    def emit(self):
        yield self.part('true')


class JSFalse(JSNode):
    def emit(self):
        yield self.part('false')


class JSNull(JSNode):
    def emit(self):
        yield self.part('null')


class JSRest(JSNode):
    def emit(self, value):
        yield self.part('...', value)


class JSTaggedTemplate(JSNode):

    def emit(self, value, func):
        text = list(delimited_multi_line(self, value, '`'))
        func = list(func.serialize())
        yield self.part(*func, *text)


class JSTemplateLiteral(JSNode):

    def emit(self, value):
        yield from delimited_multi_line(self, value, '`')


#### Ops


class JSOpIn(JSNode):
    def emit(self):
        yield self.part('in')


class JSOpAnd(JSNode):
    def emit(self):
        yield self.part('&&')


class JSOpOr(JSNode):
    def emit(self):
        yield self.part('||')


class JSOpNot(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('!')


class JSOpInstanceof(JSNode):
    def emit(self):
        yield self.part('instanceof')


class JSOpTypeof(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('typeof')


class JSOpAdd(JSNode):
    def emit(self):
        yield self.part('+')


class JSOpSub(JSNode):
    def emit(self):
        yield self.part('-')


class JSOpMult(JSNode):
    def emit(self):
        yield self.part('*')


class JSOpDiv(JSNode):
    def emit(self):
        yield self.part('/')


class JSOpMod(JSNode):
    def emit(self):
        yield self.part('%')


class JSOpRShift(JSNode):
    def emit(self):
        yield self.part('>>')


class JSOpLShift(JSNode):
    def emit(self):
        yield self.part('<<')


class JSOpBitXor(JSNode):
    def emit(self):
        yield self.part('^')


class JSOpBitAnd(JSNode):
    def emit(self):
        yield self.part('&')


class JSOpBitOr(JSNode):
    def emit(self):
        yield self.part('|')


class JSOpInvert(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('~')


class JSOpUSub(JSLeftSideUnaryOp):
    def emit(self):
        yield self.part('-')


class JSOpStrongEq(JSNode):
    def emit(self):
        yield self.part('===')


class JSOpStrongNotEq(JSNode):
    def emit(self):
        yield self.part('!==')


class JSOpLt(JSNode):
    def emit(self):
        yield self.part('<')


class JSOpLtE(JSNode):
    def emit(self):
        yield self.part('<=')


class JSOpGt(JSNode):
    def emit(self):
        yield self.part('>')


class JSOpGtE(JSNode):
    def emit(self):
        yield self.part('>=')

JSIs = JSOpStrongEq



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- code to aid transformation, it gets converted
# :Created:  mar 01 mar 2016 01:42:26 CET
# :Author:   Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#


def _in(left, right):
    from __globals__ import Array, typeof

    if isinstance(right, Array) or typeof(right) == 'string':
        return right.indexOf(left) > -1
    else:
        return left in right


def in_es6(left, right):
    from __globals__ import Array, typeof, Map, Set, WeakMap, WeakSet

    if isinstance(right, Array) or typeof(right) == 'string':
        return right.indexOf(left) > -1
    elif isinstance(right, (Map, Set, WeakMap, WeakSet)):
        return right.has(left)
    else:
        return left in right


def set_decorators(cls, props):
    from __globals__ import Function, Map, WeakMap, Object

    for p in dict(props):
        decos = props[p]
        def reducer(val, deco):
            return deco(val, cls, p)
        deco = decos.reduce(reducer, cls.prototype[p])
        if not isinstance(deco, (Function, Map, WeakMap)) and \
            isinstance(deco, Object) and (('value' in deco) or
                                          ('get' in deco)):
            del cls.prototype[p]
            Object.defineProperty(cls.prototype, p, deco)
        else:
            cls.prototype[p] = deco


def set_class_decorators(cls, decos):
    def reducer(val, deco):
        return deco(val, cls)
    return decos.reduce(reducer, cls)


def set_properties(cls, props):
    from __globals__ import Function, Map, WeakMap, Object

    for p in dict(props):
        value = props[p]
        if not isinstance(value, (Map, WeakMap)) and isinstance(value, Object) \
           and 'get' in value and isinstance(value.get, Function):
            # the following condition raises a TypeError in dukpy, why?
            # ('set' in value and isinstance(value.set, Function)):
            desc = value
        else:
            desc = {
                'value': value,
                'enumerable': False,
                'configurable': True,
                'writable': True
            }
        Object.defineProperty(cls.prototype, p, desc)


def _assert(comp, msg):
    from __globals__ import Error, Object, typeof

    def PJAssertionError(self, message):
        self.name = 'PJAssertionError'
        self.message = message or 'Custom error PJAssertionError'
        if typeof(Error.captureStackTrace) == 'function':
            Error.captureStackTrace(self, self.constructor)
        else:
            self.stack = Error(message).stack

    PJAssertionError.prototype = Object.create(Error.prototype)
    PJAssertionError.prototype.constructor = PJAssertionError

    msg = msg or 'Assertion failed.'
    if not comp:
        raise PJAssertionError(msg)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.pj -- A Python 3 to ES6 Javascript translator
# :Created:   lun 22 feb 2016 16:00:00 CET
# :Author:    Alberto Berti <alberto@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Alberto Berti
#

import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.rst'), encoding='utf-8') as f:
    CHANGES = f.read()
with open(os.path.join(here, 'version.txt'), encoding='utf-8') as f:
    VERSION = f.read().strip()

setup(
    name="javascripthon",
    version=VERSION,
    url="https://github.com/azazel75/metapensiero.pj",

    description="Javascript for refined palates:"
    " a Python 3 to ES6 Javascript translator",
    long_description=README + '\n\n' + CHANGES,

    author="Alberto Berti",
    author_email="alberto@arstecnica.it",

    license="GPLv3+",
    classifiers=[
        'Development Status :: 4 - Beta',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        ],
    keywords='JavaScript EcmaScript compilation translation transpiling babel',
    packages=['metapensiero.' + pkg
              for pkg in find_packages('src/metapensiero')],
    package_dir={'': 'src'},
    namespace_packages=['metapensiero',],
    install_requires=[
        'setuptools',
        'dukpy',
        'macropy3==1.0.4.dev2'
    ],
    extras_require={
        'dev': [
            'metapensiero.tool.bump_version',
            'readme_renderer',
        ],
        'test': [
            'pytest',
            'meta'
        ],
    },
    setup_requires=[
        'pytest-runner'
    ],
    tests_require=[
        'javascripthon[test]'
    ],
    entry_points={
        'console_scripts': ['pj=metapensiero.pj.__main__:main'],
    },
    package_data={
        'metapensiero.pj': ['data/*.js']
    },
)



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- transformation processor
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
import collections
import contextlib
import inspect
import os
import sys
import string
import textwrap

from .exceptions import TransformationError, UnsupportedSyntaxError
from .util import (rfilter, parent_of, Line, Part, obj_source,
                   body_local_names, walk_under_code_boundary)

SNIPPETS_TEMPLATE = """\
def _pj_snippets(container):
%(snippets)s
%(assignments)s
    return container

_pj = {}
_pj_snippets(_pj)

"""

ASSIGN_TEMPLATE = "    container['%(name)s'] = %(name)s"

VAR_TEMPLATE = "_pj_%s"


class TargetNode:
    """This is the common ancestor of all the JS AST nodes."""

    """The associated Python AST node"""
    py_node = None

    """The Transformer instance which is managing this one"""
    transformer = None

    """The final arguments passed to the emit method defined on the subclasses"""
    transformed_args = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.options = kwargs

    def __str__(self):
        return ''.join(
            str(x) for x in
            self.serialize())

    @classmethod
    def final(cls, *transformed_args, **options):
        tn = cls(**options)
        tn.transformed_args = transformed_args
        return tn

    def serialize(self):
        for a in self.emit(*self.transformed_args, **self.options):
            yield from a.serialize()

    def lines(self, items, *, indent=False, delim=False, name=None):
        if not isinstance(items, (tuple, list)):
            items = (items,)
        for i in self._chain(items):
            yield self.line(i, indent=indent, delim=delim, name=name)

    def line(self, item, indent=False, delim=False, name=None):
        if isinstance(item, Line):
            item.indent += int(indent)
            l = item
        elif isinstance(item, (tuple, list)):
            item = tuple(self._chain(item))
            l = Line(self, item, indent, delim, name)
        else:
            l = Line(self, item, indent, delim, name)
        return l

    def part(self, *items, name=None):
        it = tuple(self._expand(items))
        if len(it) == 1 and isinstance(it[0], Line):
            result = it[0].item
        else:
            result = Part(self, *it, name=name)
        return result

    def _expand(self, items):
        for i in items:
            if isinstance(i, TargetNode):
                yield from i.serialize()
            else:
                yield i

    def _chain(self, items):
        for i in self._expand(items):
            if inspect.isgenerator(i):
                yield from i
            else:
                yield i


class Transformer:

    enable_snippets = True
    enable_es6 = False
    enable_let = False
    enable_stage3 = False
    disable_srcmap = False

    """Used in subtransformation to remap a node on a Transformer instance to the
    AST produced by a substransform."""
    remap_to = None

    def __init__(self, py_ast_module, statements_class, snippets=True,
                 es6=False, stage3=False, remap_to=None):
        self.transformations = load_transformations(py_ast_module)
        self.statements_class = statements_class
        self.enable_snippets = snippets
        self.enable_es6 = es6
        self.enable_stage3 = stage3
        self.remap_to = remap_to
        self._init_structs()

    def _init_structs(self):
        self.snippets = set()
        self._globals = set()
        self._args_stack = []
        self._context = collections.ChainMap()
        self._warnings = []

    @property
    def ctx(self):
        return self._context

    def _push_ctx(self, **kwargs):
        self._context = self._context.new_child(kwargs)

    def _pop_ctx(self):
        self._context = self._context.parents

    @contextlib.contextmanager
    def context_for(self, py_node, **kwargs):
        if isinstance(py_node, ast.stmt):
            self._push_ctx(**kwargs)
            yield
            self._pop_ctx()
        else:
            yield

    @classmethod
    def new_from(cls, instance):
        new = cls.__new__(cls)
        new._init_structs()
        new.transformations = instance.transformations
        new.statements_class = instance.statements_class
        for k, v in vars(instance).items():
            if k.startswith('enable_'):
                setattr(new, k, v)
        return new

    def transform_code(self, ast_tree):
        """Convert the given Python AST dump into JavaScript AST."""
        from ..js_ast import JSVarStatement

        top = ast.parse(ast_tree)
        body = top.body
        self._args_stack.clear()

        self.node_parent_map = build_node_parent_map(top)

        local_vars = body_local_names(body)
        self.ctx['vars'] = local_vars
        result = self.statements_class(*body)
        self._finalize_target_node(result)

        local_vars = list(local_vars - self._globals)
        if len(local_vars) > 0:
            local_vars.sort()
            vars = JSVarStatement(local_vars,
                                  [None] * len(local_vars))
            self._finalize_target_node(vars)
            result.transformed_args.insert(0, vars)

        self.node_parent_map = None

        return result

    def parent_of(self, node):
        """Return the parent of the given node."""
        return self.node_parent_map.get(node)

    def parents(self, node, stop_at=None):
        """Return all the parents possibly up an instance of `stop_at` class."""
        parent = self.node_parent_map.get(node)
        while parent:
            yield parent
            if stop_at and isinstance(parent, stop_at):
                break
            parent = self.node_parent_map.get(parent)

    def find_parent(self, node, *classes):
        """Retrieve the first parent of the given AST `node` that is an instance
        of the given `classes`."""
        parent = self.parent_of(node)
        if parent is not None:
            if isinstance(parent, classes):
                return parent
            else:
                return self.find_parent(parent, *classes)

    def find_child(self, node, cls):
        """Find any child of `node` that is an instance of `cls`. The walk will
        not go down into other code blocks."""
        if not isinstance(node, (tuple, list, set)):
            node = (node)
        for n in node:
            for c in walk_under_code_boundary(n):
                if isinstance(c, cls):
                    yield c

    def has_child(self, node, cls):
        """Return true if `node` has any child that is an instance of `cls`."""
        wanted = tuple(self.find_child(node, cls))
        return len(wanted) > 0

    def new_name(self):
        """Generate a new name to use in statements."""
        ix = self.ctx.setdefault('gen_name_ix', -1)
        ix += 1
        self.ctx['gen_name_ix'] = ix
        if ix > len(string.ascii_letters):
            raise TransformationError("Reached maximum index for "
                                      "auto-generated variable names")
        return VAR_TEMPLATE % string.ascii_letters[ix]

    def add_snippet(self, func):
        """Add a function to the snippets."""
        self.snippets.add(func)

    def _transform_node(self, in_node):
        """Transform a Python AST node to a JS AST node."""

        if isinstance(in_node, list) or isinstance(in_node, tuple):
            res = [self._transform_node(child) for child in in_node]

        elif isinstance(in_node, ast.AST):
            # prepare a context for the transformation if it's a
            # statement; it's used for example by try...catch stmts to
            # give hints to raise
            with self.context_for(in_node):
                # transformations can come in tuples or lists, take the
                # first one
                for transformation in self.transformations.get(
                        in_node.__class__.__name__, []):
                    out_node = transformation(self, in_node)
                    if out_node is not None:
                        self._finalize_target_node(out_node, py_node=in_node)
                        res = out_node
                        break
                else:
                    raise TransformationError(in_node,
                                              "No transformation for the node")

        elif isinstance(in_node, TargetNode):
            self._finalize_target_node(in_node)
            res = in_node
        else:
            # e.g. an integer
            res = in_node
        return res

    def _finalize_target_node(self, tnode, py_node=None):
        tnode.py_node = self.remap_to or py_node or tnode.py_node
        tnode.transformer = self
        if tnode.transformed_args is None:
            tnode.transformed_args = targs = []
            args = collections.deque(tnode.args)
            self._args_stack.append(args)
            while args:
                arg = args.popleft()
                targs.append(self._transform_node(arg))
            self._args_stack.pop()

    def transform_snippets(self):
        snippets = tuple(sorted(self.snippets, key=lambda e: e.__name__))
        srcs = [obj_source(s) for s in snippets]
        src = textwrap.indent('\n'.join(srcs), ' ' * 4)
        names = [s.__name__ for s in snippets]
        assign_src = '\n'.join([ASSIGN_TEMPLATE % {'name': n} for n in names])
        trans_src = SNIPPETS_TEMPLATE % {
            'snippets': src,
            'assignments': assign_src
        }
        t = self.new_from(self)
        t.snippets = None
        t.enable_snippets = False
        t.disable_srcmap = True
        return t.transform_code(trans_src)

    def add_globals(self, *items):
        self._globals |= set(items)

    def _guard(self, test, node, desc):
        if not test:
            raise TransformationError(node, desc)

    def es6_guard(self, node, desc):
        """Raise an exception if es6 isn't enabled."""
        self._guard(self.enable_es6, node, desc)

    def next_args(self):
        return self._args_stack[-1]

    def unsupported(self, py_node, cond, desc):
        """Raise an exception if `cond` is ``False``."""
        if cond:
            raise UnsupportedSyntaxError(py_node, desc)

    def warn(self, py_node, msg):
        """Append the given message to the warnings."""
        self._warnings.append((py_node, msg))

    def subtransform(self, obj, remap_to=None):
        """Transform a piece of code, either a python object or a string. This
        is done in a new ``Transformer`` with a configuration similar to the
        calling instance."""
        if isinstance(obj, str):
            src = textwrap.dedent(obj)
        else:
            src = obj_source(obj)
        t = self.new_from(self)
        t.remap_to = remap_to
        t.snippets = None
        t.enable_snippets = False
        return t.transform_code(src)

    def stage3_guard(self, node, desc):
        """Raise an exception if stage3 isn't enabled."""
        self._guard(self.enable_stage3, node, desc)


#### Helpers

def python_ast_names():
    #LATER: do this properly
    return rfilter(r'[A-Z][a-zA-Z]+', dir(ast))


def load_transformations(py_ast_module):
    # transformationsDict = {
    #     'NodeName': [...transformation functions...]
    # }
    d = {}
    ast_names = list(python_ast_names())
    filenames = rfilter(
        r'^[^.]+\.py$',
        os.listdir(parent_of(py_ast_module.__file__)))
    for filename in filenames:
        if filename != '__init__.py':
            mod_name = 'metapensiero.pj.transformations.%s' % \
                       filename.split('.')[0]
            __import__(mod_name)
            mod = sys.modules[mod_name]
            for name in dir(mod):
                if name in ast_names:
                    assert name not in d
                    value = getattr(mod, name)
                    if not isinstance(value, list) or isinstance(value, tuple):
                        value = [value]
                    d[name] = value
    return d


def build_node_parent_map(top):

    node_parent_map = {}

    def _process_node(node):
        for k in node._fields:
            x = getattr(node, k)
            if not (isinstance(x, list) or isinstance(x, tuple)):
                x = [x]
            for y in x:
                if isinstance(y, ast.AST):
                    node_parent_map[y] = node
                    _process_node(y)

    _process_node(top)

    return node_parent_map



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- test utilities
# :Created:  mer 09 nov 2016 20:32:09 CET
# :Author:   Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast
import inspect
import textwrap

from . import transformations
from .js_ast import JSStatements
from .processor.transforming import Transformer


def ast_object(obj):
    """Convert a Python object to its AST representation."""
    src = inspect.getsource(obj)
    # ast wraps always into a module
    node = ast.parse(textwrap.dedent(src)).body[0]
    return node


def ast_dump_object(obj, first_stmt_only=False):
    """Convert a Python object to its AST representation and the serialization of
    this to a string, optionally serializing the first element in its body.
    """
    from meta.asttools import str_ast
    node = ast_object(obj)
    if first_stmt_only:
        node = node.body[0]
    return node, str_ast(node)


def ast_object_to_js(obj, es6=False):
    """Convert a Python object to JS using pj, optionally transforming with ES6
    features enabled.
    """
    src = inspect.getsource(obj)
    node = ast.parse(textwrap.dedent(src))
    t = Transformer(transformations, JSStatements, es6=es6)
    return t.transform_code(node)


def ast_dump_file(fname):
    """Dump an entire file."""
    with open(fname) as f:
        return ast_dumps(f.read(), filename=fname)


def ast_dumps(input, filename='', first_stmt_only=False):
    """AST dump a string."""
    try:
        from meta.asttools import str_ast
    except ImportError:
        str_ast = None
    node = ast.parse(input, filename=filename)
    if first_stmt_only:
        node = node.body[0]
    if str_ast:
        dump = str_ast(node)
    else:
        dump = ""
    return node, dump



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- for loops transformations
# :Created:  ven 26 feb 2016 15:17:49 CET
# :Authors:  Andrew Schaaf <andrew@andrewschaaf.com>,
#            Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import ast

from ..js_ast import (
    JSAssignmentExpression,
    JSAttribute,
    JSAugAssignStatement,
    JSBinOp,
    JSCall,
    JSExpressionStatement,
    JSForStatement,
    JSForeachStatement,
    JSForofStatement,
    JSIfStatement,
    JSName,
    JSNum,
    JSOpAdd,
    JSOpLt,
    JSStatements,
    JSSubscript,
    JSVarStatement,
)


def For_range(t, x):
    """Special conversion for ``for name in range(n)``, which detects
    ``range()`` calls and converts the statement to:

    .. code:: javascript

      for (var name = 0, bound = bound; name < bound; name++) {
          // ...
      }

    """
    if (isinstance(x.target, ast.Name) and
        isinstance(x.iter, ast.Call) and
        isinstance(x.iter.func, ast.Name) and
        x.iter.func.id == 'range' and
        1 <= len(x.iter.args) < 4) and (not x.orelse):

        name = x.target
        body = x.body
        if len(x.iter.args) == 1:
            start = JSNum(0)
            bound = x.iter.args[0]
            step = JSNum(1)
        elif len(x.iter.args) == 2:
            start = x.iter.args[0]
            bound = x.iter.args[1]
            step = JSNum(1)
        else:
            start = x.iter.args[0]
            bound = x.iter.args[1]
            step = x.iter.args[2]

        bound_name = t.new_name()

        return JSForStatement(
            JSVarStatement(
                [name.id, bound_name],
                [start, bound]),
            JSBinOp(JSName(name.id), JSOpLt(), JSName(bound_name)),
            JSAugAssignStatement(
                JSName(name.id), JSOpAdd(), step),
            body
        )


def For_dict(t, x):
    """Special ``for name in dict(expr)`` statement translation.

    It detects the ``dict()`` call and converts it to:

    .. code:: javascript

      var dict_ = expr;
      for (var name in dict_) {
          if (dict_.hasOwnProperty(name)) {
          // ...
          }
      }
    """
    if (isinstance(x.iter, ast.Call) and
        isinstance(x.iter.func, ast.Name) and
        x.iter.func.id == 'dict' and
        len(x.iter.args) <= 2) and (not x.orelse):

        t.unsupported(x, not isinstance(x.target, ast.Name),
                      "Target must be a name")

        name = x.target
        expr = x.iter.args[0]
        body = x.body

        dict_ = t.new_name()

        # if not ``dict(foo, True)`` filter out inherited values
        if not (len(x.iter.args) == 2 and
            isinstance(x.iter.args[1], ast.NameConstant) and
            x.iter.args[1].value):
            body = [
                    JSIfStatement(
                        JSCall(
                            JSAttribute(JSName(dict_), 'hasOwnProperty'),
                            [JSName(name.id)]
                        ),
                        body, None
                    )
                ]
        # set the incoming py_node for the sourcemap
        loop = JSForeachStatement(
            name.id,
            JSName(dict_),
            body
        )
        loop.py_node = x

        return JSStatements(
            JSVarStatement([dict_], [expr], unmovable=True),
            loop
        )


def For_iterable(t, x):
    """Special ``for name in iterable(expr)`` statement translation.

    It detects the ``iterable()`` call and converts it to:

    .. code:: javascript

      var __iterable = expr;
      for (var name of __iterable) {
          ...
      }
    """
    if (isinstance(x.iter, ast.Call) and
        isinstance(x.iter.func, ast.Name) and
        x.iter.func.id == 'iterable' and
        len(x.iter.args) == 1) and (not x.orelse):
        t.es6_guard(x, 'for...of statement requires ES6')

        expr = x.iter.args[0]
        body = x.body
        target = x.target

        # set the incoming py_node for the sourcemap
        return JSForofStatement(
            target,
            expr,
            body,
        )


def For_default(t, x):
    """Assumes that the iteration is over a list.

    Converts something like:

    .. code:: python

      for name in expr:
          #body...

    to:

    .. code:: javascript

      for(var name, arr=expr, ix=0, length=arr.length; ix < length: ix++) {
          name = arr[ix];
          //body ...
      }

    """

    t.unsupported(x, not isinstance(x.target, ast.Name), "Target must be a name,"
                  " Are you sure is only one?")

    name = x.target
    expr = x.iter
    body = x.body

    arr = t.new_name()
    length = t.new_name()
    ix = t.new_name()

    return JSForStatement(
        JSVarStatement(
            [name.id, ix, arr, length],
            [None, JSNum(0), expr,
             JSAttribute(JSName(arr), 'length')],
            unmovable=True
        ),
        JSBinOp(
            JSName(ix),
            JSOpLt(),
            JSName(length)),
        JSExpressionStatement(
            JSAugAssignStatement(
                JSName(ix),
                JSOpAdd(),
                JSNum(1))),
        [
            JSExpressionStatement(
                JSAssignmentExpression(
                    JSName(name.id),
                    JSSubscript(
                        JSName(arr),
                        JSName(ix))))
        ] + body)


For = [For_range, For_dict, For_iterable, For_default]



# -*- coding: utf-8 -*-
# :Project:  metapensiero.pj -- commandline interface
# :Created:  mar 01 mar 2016 19:33:21 CET
# :Author:   Alberto Berti <alberto@metapensiero.it>
# :License:  GNU General Public License version 3 or later
#

import argparse
from collections import deque
import logging
from pathlib import Path
import sys

from . import api

log = logging.getLogger(__name__)


class UnsupportedPythonError(Exception):
    """Exception raised if the running interpreter version is
    unsupported.
    """


parser = argparse.ArgumentParser(
    description="A Python 3.5+ to ES6 JavaScript compiler",
    prog='pj'
)
parser.add_argument('files', metavar='file', type=str, nargs='*',
                    help="Python source file(s) or directory(ies) "
                    "to convert. When it is a directory it will be "
                    "converted recursively")
parser.add_argument('--disable-es6', dest='es6', action='store_false',
                    default=True,
                    help="Disable ES6 features during conversion"
                    " (Ignored if --es5 is specified)")
parser.add_argument('--disable-stage3', dest='stage3', action='store_false',
                    default=True,
                    help="Disable ES7 stage3 features during conversion")
parser.add_argument('-5', '--es5', dest='es5', action='store_true',
                    help="Also transpile to ES5 using BabelJS.")
parser.add_argument('--transform-runtime', action='store_true', dest='truntime',
                    help="Add transform runtime as plugin during transpile")
parser.add_argument('-o', '--output', type=str,
                    help="Output file/directory where to save the generated "
                    "code")
parser.add_argument('-d', '--debug', action='store_true',
                    help="Enable error reporting")
parser.add_argument('--pdb', action='store_true',
                    help="Enter post-mortem debug when an error occurs")
parser.add_argument('-s', '--string', type=str,
                    help="Convert a string, useful for small snippets. If the string"
                    " is '-' will be read from the standard input.")
parser.add_argument('-e', '--eval', action='store_true',
                    help="Evaluate the string supplied with the -s  using the"
                    " embedded interpreter and return the last result. This will "
                    "convert the input string with all the extensions enabled "
                    "(comparable to adding the '-5' option) and so it will take"
                    " some time because of BabelJS load times.")
parser.add_argument('--dump-ast', action='store_true',
                    help="Dump the Python AST. You need to have the package"
                    " metapensiero.pj[test] installed")
parser.add_argument('--inline-map', action='store_true',
                    help="Save the source-map inline instead of in an additional"
                    " file, useful when transpiling with BabelJS externally "
                    "but without access to the cli. Ignored "
                    "when transpiling.")



class Reporter:
    def __init__(self, fout=None, ferr=None):
        self.fout = fout or sys.stdout
        self.ferr = ferr or sys.stderr

    def print_err(self, *args, **kwargs):
        kwargs['file'] = self.ferr
        print(*args, **kwargs)

    def print(self, *args, **kwargs):
        kwargs['file'] = self.fout
        print(*args, **kwargs)


def transform(src_fname, dst_fname=None, transpile=False, enable_es6=False,
              enable_stage3=False, **kw):
    if transpile:
        kw.pop('inline_map', None)
        api.transpile_py_file(src_fname, dst_fname,
                              enable_stage3=enable_stage3,
                              **kw)
    else:
        kw.pop('truntime', None)
        api.translate_file(src_fname, dst_fname, enable_es6=enable_es6,
                           enable_stage3=enable_stage3, **kw)


def transform_string(input, transpile=False, enable_es6=False,
                     enable_stage3=False, **kw):
    if transpile:
        res, src_map = api.transpile_pys(input, enable_stage3=enable_stage3)
    else:
        res, src_map = api.translates(input, enable_es6=enable_es6,
                                      enable_stage3=enable_stage3)
    if kw.get('inline_map', False):
        res += api._inline_src_map(src_map)
    return res


def check_interpreter_supported():
    if sys.version_info < (3, 5):
        raise UnsupportedPythonError('JavaScripthon needs at least'
                                     ' Python 3.5 to run')


def main(args=None, fout=None, ferr=None):
    result = 0
    rep = Reporter(fout, ferr)
    args = parser.parse_args(args)
    freeargs = {
        'truntime': args.truntime,
        'inline_map': args.inline_map
    }
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)
        log.debug('Log started')
    if not (args.files or args.string):
        rep.print_err("Error: You have to supply either a string with -s or a "
                      "filename")
        result = 3
    if args.eval and not args.string:
        rep.print_err("Error: You have to supply a string with -s when using "
                      "evaluation")
    elif args.string:
        check_interpreter_supported()
        input = args.string
        if input == '-':
            input = sys.stdin.read()
        if args.dump_ast:
            from .testing import ast_dumps
            rep.print(ast_dumps(input)[1])
        if args.eval:
            es6 = es5 = stage3 = True
        else:
            es5 = args.es5
            es6 = args.es6
            stage3 = args.stage3
        try:
            res = transform_string(input, es5, es6, stage3,
                                   **freeargs)
            if args.eval:
                res = api.evaljs(res, load_es6_polyfill=True)
            rep.print(res)
        except Exception as e:
            if args.pdb:
                import pdb
                pdb.post_mortem(e.__traceback__)
            elif args.debug:
                raise
            else:
                error = "%s: %s" % (e.__class__.__name__, e)
                rep.print_err("An error occurred while compiling source "
                              "from string")
                rep.print_err(error)
            result = 1
    elif args.output and len(args.files) > 1:
        rep.print_err("Error: only one source file is allowed when "
                      "--output is specified.")
        result = 2
    else:
        try:
            check_interpreter_supported()
            for fname in args.files:
                src = Path(fname)
                if not src.exists():
                    rep.print_err("Skipping non existent file '%s'" % src)
                    continue
                if args.output:
                    dst = Path(args.output)
                else:
                    dst = None
                if src.is_dir():
                    if dst and src != dst:
                        if dst.exists() and not dst.is_dir():
                            rep.print_err("Source is a directory but output exists "
                                          "and it isn't")
                            result = 1
                            break
                        if not dst.exists():
                            dst.mkdir()
                    src_root = src
                    dst_root = dst
                    src_dirs = deque([src_root])
                    while len(src_dirs) > 0:
                        sdir = src_dirs.popleft()
                        if dst_root:
                            ddir = dst_root / sdir.relative_to(src_root)
                            if not ddir.exists():
                                ddir.mkdir()
                        else:
                            ddir = None
                        for spath in sdir.iterdir():
                            if spath.name in ('__pycache__', '__init__.py'):
                                continue
                            elif spath.is_dir():
                                src_dirs.append(spath)
                                continue
                            elif spath.suffix == '.py':
                                try:
                                    transform(
                                        str(spath),
                                        str(ddir) if ddir else None,
                                        args.es5,
                                        args.es6,
                                        args.stage3,
                                        **freeargs
                                    )
                                    rep.print("Compiled file %s" % spath)
                                except Exception as e:
                                    e.src_fname = spath
                                    raise
                else:
                    try:
                        transform(fname, args.output, args.es5, args.es6,
                                  args.stage3, **freeargs)
                        rep.print("Compiled file %s" % fname)
                    except Exception as e:
                        e.src_fname = fname
                        raise
        except Exception as e:
            if args.pdb:
                import pdb
                pdb.post_mortem(e.__traceback__)
            elif args.debug:
                raise
            else:
                src_fname = getattr(e, 'src_fname', None)
                error = "%s: %s" % (e.__class__.__name__, e)
                if src_fname:
                    rep.print_err("An error occurred while compiling source "
                                  "file '%s'" % src_fname)
                else:
                    rep.print_err("An error occured during processing.")
                rep.print_err(error)
            result = 1
    sys.exit(result)


if __name__ == '__main__':
    main()
